use crate::{
    disk::{
        partition::{Partition, PartitionStatus},
        Disk,
    },
    fat::{direntry::NodeRef, Fat},
};

pub struct Analyzer {
    disk: Disk,
    filesystem: Fat,
}

impl Analyzer {
    pub fn load(file: &str) -> Result<Self, std::io::Error> {
        let mut disk = Disk::load(file)?;
        let partition = Self::get_boot_partition(&disk).expect("Oops!");
        let filesystem = Fat::load(&mut disk, &partition)?;
        Ok(Analyzer { disk, filesystem })
    }

    pub fn print_metadata(&mut self) {
        let rootdir_ref = NodeRef::new(self.filesystem.root_directory, &self.filesystem.arena);

        println!("Disk metadata according to VHD footer\n-------------------------------------\n");
        println!("{}", self.disk.get_metadata());
        println!("Partition table according to the MBR\n------------------------------------\n");
        println!("{}", self.disk.partition_table);
        println!("Filesystem information\n-----------------------------------------------\n");
        println!("{}", self.filesystem);
        println!("Root directory information\n----------------------------------\n");
        println!("{}", rootdir_ref);
    }

    fn get_boot_partition(disk: &Disk) -> Option<Partition> {
        let mut partitions: Vec<Partition> = Vec::with_capacity(4);
        for partition in &disk.partition_table.partitions {
            match partition {
                Some(part) => partitions.push(part.clone()),
                _ => println!("Nothing found"),
            }
        }
        if partitions.is_empty() {
            return None;
        }

        for partition in partitions {
            if partition.get_status() == PartitionStatus::Active {
                return Some(partition);
            }
        }
        None
    }
}
