use std::str::FromStr;

use crate::{
    disk::{
        partition::{Partition, PartitionType},
        Disk,
    },
    downloader::ZipDownloader,
    fat::{
        direntry::{DirEntry, DirEntryType, NodeId},
        vbr::Vbr,
        Fat,
    },
    manifest::Manifest,
    os::OperatingSystem,
};

#[derive(Debug)]
pub struct DiskBuilder {
    filesystem: Fat,
    operating_system: OperatingSystem,
    disk: Disk,
    partition: Partition,
    manifest: Manifest,
}

impl DiskBuilder {
    /// Create a new builder from a manifest
    pub fn new(manifest: Manifest) -> Result<Self, std::io::Error> {
        // Gather what we need from the manifest and instantiate structs as needed
        let mut disk = Disk::new(manifest.disktype.as_str(), manifest.os.as_str())?;
        let operating_system = OperatingSystem::from_str(manifest.os.as_str()).unwrap();
        disk.attach(manifest.diskname.as_str())?;
        disk.set_geometry(manifest.disksize)?;
        let root_partition =
            Partition::fill_disk(&disk, Some(Self::derive_partition_type(&manifest)))?;
        disk.partition_table
            .add_partition(root_partition.clone(), Some(0))?;
        let filesystem = Fat::new(&disk, &root_partition, Some(operating_system))?;

        // Return the completed disk builder instance
        Ok(DiskBuilder {
            disk,
            filesystem,
            partition: root_partition,
            operating_system,
            manifest,
        })
    }

    /// Copy a file from a downloaded ZIP to the VHD
    pub fn copy_file(
        &mut self,
        downloader: &ZipDownloader,
        filename: &str,
    ) -> Result<(), std::io::Error> {
        if let Some(source_file) = downloader.get_file_by_name(filename) {
            self.filesystem
                .mkfile(None, source_file.get_name(), false, source_file.get_bytes())?;
            Ok(())
        } else {
            // If the file was not found, nothing essentially bad happens. Just continue and do nothing for now.
            Ok(())
        }
    }

    /// Copy a file from a downloaded ZIP to the VHD
    pub fn copy_sysfile(
        &mut self,
        downloader: &ZipDownloader,
        filename: &str,
    ) -> Result<(), std::io::Error> {
        if let Some(source_file) = downloader.get_file_by_name(filename) {
            self.filesystem
                .mkfile(None, source_file.get_name(), true, source_file.get_bytes())?;
            Ok(())
        } else {
            // If the file was not found, nothing essentially bad happens. Just continue and do nothing for now.
            Ok(())
        }
    }

    /// Writes the contents of a `DirEntry` to the disk in clusters.
    ///
    /// This function takes the data from a `DirEntry` and writes it to the disk,
    /// segmenting the data into clusters based on the filesystem's cluster size.
    /// Each segment of data is written to a specific disk cluster corresponding
    /// to pre-allocated cluster addresses within the `DirEntry`.
    ///
    /// # Parameters
    /// - `file`: A reference to the `DirEntry` containing the data to be written.
    ///   The `DirEntry` must contain both the data (as a byte slice) and a list
    ///   of allocated cluster addresses where the data will be written.
    ///
    /// # Returns
    /// This function returns `Ok(())` if the data is successfully written to all
    /// specified clusters. It returns an `Err(std::io::Error)` if there is a mismatch
    /// in the number of allocated clusters and the number of data chunks, or if
    /// any call to `write_cluster` fails.
    ///
    /// # Errors
    /// This function can error in the following scenarios:
    /// - Returns an error if called on a directory entry type other than `File` or `SysFile`.
    /// - If the number of clusters pre-allocated in `file.allocated_clusters` does
    ///   not match the number of chunks into which the data has been divided. This
    ///   situation triggers an error with `std::io::ErrorKind::Other`.
    /// - If writing any cluster fails, the underlying `write_cluster` function
    ///   may also return an `std::io::Error` describing the specific problem encountered
    ///   during the write operation.
    ///
    /// # Note
    /// This function assumes that `file.get_data()` returns a slice of all data to be written
    /// and that `file.allocated_clusters` contains a valid list of disk addresses. It also
    /// assumes that each address in `file.allocated_clusters` can hold a cluster of data
    /// of size defined by `filesystem.get_sectors_per_cluster() * 512`.
    pub fn write_entry_to_storage(&mut self, entry_id: NodeId) -> Result<(), std::io::Error> {
        let cluster_size = self.get_cluster_size();
        let entry = self.filesystem.arena.get_node(entry_id).ok_or_else(|| {
            std::io::Error::new(std::io::ErrorKind::NotFound, "Entry node not found")
        })?;
        let entry = &entry.value.clone();

        let bytes = match entry.entry_type {
            DirEntryType::File => entry.get_data().to_vec(),
            DirEntryType::SysFile => entry.get_data().to_vec(),
            DirEntryType::Directory => entry.children_bytes(&self.filesystem.arena)?,
        };

        if bytes.is_empty() {
            return Ok(());
        }

        let chunks = bytes.chunks(cluster_size).collect::<Vec<_>>(); // Collect into Vec
        let addresses = &entry.allocated_clusters; // Use a reference

        // Ensure the number of allocated clusters matches the number of chunks
        if entry.entry_type == DirEntryType::Directory && (entry.directory_name() == "." || entry.directory_name() == "..") {
            return Ok(());
        }

        // Iterate over addresses and chunks to write data directly
        for (address, chunk) in addresses.iter().zip(chunks) {
            // Call write_cluster using mutable filesystem
            self.write_cluster(*address, chunk)?;
        }

        Ok(())
    }

    fn get_cluster_size(&self) -> usize {
        self.filesystem.get_sectors_per_cluster() * 512
    }

    /// Writes the data of a cluster to the disk.
    ///
    /// This method takes the data of a cluster, represented as a `&[u8]`, and writes it to the disk.
    /// If the provided data is shorter than the size of a cluster, it is padded with zeroes.
    /// Each cluster is divided into sectors, and the data is written sector by sector.
    ///
    /// # Arguments
    ///
    /// * `data`: The data of the cluster to be written to the disk.
    /// * `cluster_address`: The address of the cluster on the disk where the data will be written.
    ///
    /// # Errors
    ///
    /// This method returns an `std::io::Error` if there is an I/O error during writing.
    pub fn write_cluster(
        &mut self,
        cluster_address: u32,
        data: &[u8],
    ) -> Result<(), std::io::Error> {
        let bytes_per_cluster = self.filesystem.get_sectors_per_cluster() * 512;
        let mut bytes = data.to_vec();

        // If our data won't fill the cluster, pad it up to the right size with zero values.
        if bytes.len() < bytes_per_cluster {
            bytes.resize(bytes_per_cluster, 0);
        }

        // Calculate the base offset to figure out where the cluster should live on the disk
        let mut sector_address = (cluster_address as usize
            * self.filesystem.get_sectors_per_cluster()
            + self.filesystem.get_first_data_sector() as usize)
            - 2 * self.filesystem.get_sectors_per_cluster();

        for chunk in bytes.chunks_exact(512) {
            let mut chunk_array = [0u8; 512];
            chunk_array.copy_from_slice(chunk);
            self.partition
                .write_sector(&mut self.disk, sector_address as u32, chunk_array)?;
            sector_address += 1;
        }
        Ok(())
    }

    /// Writes the allocation tables to the disk for the specified partition.
    pub fn write_allocation_tables(&mut self) -> Result<(), std::io::Error> {
        let allocation_table_bytes = self
            .filesystem
            .allocation_table
            .as_bytes(self.filesystem.fat_type)?;
        let mut sectors: Vec<[u8; 512]> = Vec::new();

        let mut current_chunk: [u8; 512] = [0; 512];
        let mut current_index = 0;

        for &byte in allocation_table_bytes.iter() {
            current_chunk[current_index] = byte;
            current_index += 1;

            // If the current sector is full, push it to the sectors vector and reset
            if current_index == 512 {
                sectors.push(current_chunk);
                current_chunk = [0; 512];
                current_index = 0;
            }
        }

        // If there are remaining bytes in the last chunk, fill the rest with zeros and push it
        if current_index > 0 {
            sectors.push(current_chunk);
        }

        let mut sector_offset = self.filesystem.reserved_sector_count;
        let mut counter = 0;
        while counter < self.filesystem.fat_copies {
            for (index, sector) in sectors.iter().enumerate() {
                self.partition.write_sector(
                    &mut self.disk,
                    sector_offset as u32 + index as u32,
                    *sector,
                )?;
            }
            counter += 1;
            sector_offset += self.filesystem.sectors_per_fat;
        }

        Ok(())
    }

    /// Writes directory entries from the root directory into the disk sectors.
    ///
    /// This function serializes the directory entries from the root directory into bytes and writes
    /// them into the disk sectors starting from a calculated sector offset. The entries are written
    /// in 512-byte sectors. If the last set of bytes doesn't fill an entire sector, the remainder
    /// is zero-padded and written as the final sector.
    ///
    /// ## Parameters
    /// - `&mut self`: Mutable reference to the current instance of the object that includes
    ///   filesystem and partition information. This is necessary for accessing the filesystem layout
    ///   and performing write operations.
    ///
    /// ## Returns
    /// - `Result<(), std::io::Error>`: The function returns `Ok(())` if all sectors are written
    ///   successfully to the disk. If there is an error during any of the write operations, an
    ///   `Err` containing the `std::io::Error` is returned.
    ///
    /// ## Errors
    /// - This function will return an `Err` if there is an issue with writing to the disk.
    /// - Errors can include but are not limited to I/O errors like disk write failures.
    ///
    pub fn write_dirs(&mut self) -> Result<(), std::io::Error> {
        // Assuming root_directory is a NodeId pointing to the root directory in the Arena
        let root_id = self.filesystem.root_directory;

        // Recursively write the root directory and all its subdirectories and files
        self.write_directory_recursive(root_id)
    }

    /// Recursively writes a directory and its contents.
    fn write_directory_recursive(&mut self, dir_id: NodeId) -> Result<(), std::io::Error> {
        // Write the directory itself to storage
        self.write_entry_to_storage(dir_id)?;

        // Get the directory node from the arena
        let child_ids = {
            let dir_node = self.filesystem.arena.get_node(dir_id).ok_or_else(|| {
                std::io::Error::new(std::io::ErrorKind::NotFound, "Directory node not found")
            })?;

            // Clone child IDs to avoid borrowing issues during recursion
            dir_node.value.children.clone()
        };

        // Now iterate over each child ID and recursively write directories or files
        for &child_id in &child_ids {
            self.write_directory_recursive(child_id)?;
        }

        Ok(())
    }
    fn write_root_dir(&mut self) -> Result<(), std::io::Error> {
        // Calculate the starting sector for the root directory.
        let sector_offset: u32 = self.filesystem.reserved_sector_count as u32
            + (self.filesystem.fat_copies as u32 * self.filesystem.sectors_per_fat as u32);

        // Assuming root_directory is a NodeId pointing to the root directory in the Arena
        let root_id = self.filesystem.root_directory;
        let root_node = self.filesystem.arena.get_node(root_id).ok_or_else(|| {
            std::io::Error::new(std::io::ErrorKind::NotFound, "Root directory not found")
        })?;

        // Serialize children bytes from the root directory
        let bytes = self.serialize_directory_children(&root_node.value)?;

        // Prepare to write bytes to sectors
        let mut sectors: Vec<[u8; 512]> = Vec::new();
        let mut current_chunk: [u8; 512] = [0; 512];
        let mut current_index = 0;

        // Write the bytes off to disk
        for &byte in bytes.iter() {
            current_chunk[current_index] = byte;
            current_index += 1;

            // Got a full sector? Push it onto the vec.
            if current_index == 512 {
                sectors.push(current_chunk);
                current_chunk = [0; 512]; // Reset the chunk for new data
                current_index = 0;
            }
        }

        // Check if there is a leftover partial sector and push it if there is.
        if current_index != 0 {
            sectors.push(current_chunk);
        }

        // Write sectors to disk
        for (index, sector) in sectors.iter().enumerate() {
            self.partition
                .write_sector(&mut self.disk, sector_offset + index as u32, *sector)?;
        }

        Ok(())
    }

    /// Serialize children bytes for a given directory entry
    fn serialize_directory_children(
        &self,
        directory: &DirEntry,
    ) -> Result<Vec<u8>, std::io::Error> {
        let mut bytes = Vec::new();
        for &child_id in &directory.children {
            let child = self.filesystem.arena.get_node(child_id).ok_or_else(|| {
                std::io::Error::new(std::io::ErrorKind::NotFound, "Child directory not found")
            })?;
            bytes.extend(child.value.as_bytes()?); // Assuming DirEntry has an as_bytes method
        }
        Ok(bytes)
    }

    /// The builder is the struct that does all manner of validation before trying to
    /// create a Disk so that what we create will make sense.
    pub fn build(&mut self) -> Result<(), std::io::Error> {
        let vbr = Vbr::from_os(&self.operating_system, &self.filesystem);

        self.disk.write_os_bootcode()?;
        self.disk.write_partition_table()?;
        self.disk.write_boot_signature()?;
        self.disk.write_sector(vbr.as_bytes(), 1)?;

        // Instantiate a downloader specifically for the operating system
        let osdownloader = ZipDownloader::from_url(
            &self.operating_system.get_download_url(),
            Some(&format!(
                "Downloading {}",
                &self.operating_system.get_friendlyname()
            )),
        )?;

        // Install the required components for a bootable operating system
        self.copy_sysfile(&osdownloader, &self.operating_system.get_iosys())?;
        self.copy_sysfile(&osdownloader, &self.operating_system.get_msdossys())?;
        self.copy_file(&osdownloader, &self.operating_system.get_commandcom())?;

        // Instantiate as many downloaders as we need for the application's distfiles
        let mut gamedownloaders: Vec<ZipDownloader> =
            Vec::with_capacity(self.manifest.application.len());

        for distfile in &self.manifest.application {
            let downloader = ZipDownloader::from_url(
                &distfile.get_url(),
                Some(&format!("Downloading {}", &distfile.get_label())),
            )?;
            gamedownloaders.push(downloader);
        }

        // Process all files in the game ZIP
        for gamedownloader in gamedownloaders {
            for file in gamedownloader.get_files() {
                self.copy_file(&gamedownloader, file.get_name())?;
            }
        }

        // REMOVE THIS
        self.filesystem.mkdir(None, "DOS")?;
        // Last action: write the allocation table copies to the disk
        self.write_root_dir().unwrap();
        self.write_dirs()?;
        self.write_allocation_tables()?;
        self.disk.commit_storage()?;
        Ok(())
    }

    /// Read the Manifest struct to either get the partition type directly from it if specified,
    /// or make an educated guess as to what would be optimal given the OS and the properties of
    /// the disk itself.
    fn derive_partition_type(_manifest: &Manifest) -> PartitionType {
        // As soon as we support more OS'es, this will become a *LOT* more complex.
        // For now we only support FAT12MBR.
        PartitionType::Fat12MBR
    }
}
