use crate::os::OperatingSystem;

use self::vhd::{Vhd, VhdDisktype};
use self::{chs::CHS, partitiontable::PartitionTable};
use std::collections::BTreeMap;
use std::fs::{File, OpenOptions};
use std::io::{Read, Seek, SeekFrom, Write};
use std::str::FromStr;

pub mod chs;
pub mod partition;
pub mod partitiontable;
pub mod vhd;

#[derive(Debug)]
pub struct Disk {
    file: Option<File>,
    pub partition_table: PartitionTable,
    metadata: Vhd,
    disktype: DiskType,
    storage: BTreeMap<u32, [u8; 512]>,
    os: OperatingSystem,
    dirty: bool,
}

#[derive(PartialEq, Debug)]
pub enum DiskType {
    PCXT, // Ancient PC XT MFM disk type
    PCAT, // Slightly less ancient, somewhat bigger
    IDE,  // Everything else up to 504MB
    UNKNOWN,
}

impl Default for Disk {
    fn default() -> Self {
        Self::new("ide", "ibmdos200").unwrap()
    }
}

impl Disk {
    /// Generate a new Disk instance
    pub fn new(disktype: &str, os: &str) -> Result<Self, std::io::Error> {
        let disktype_lower = disktype.to_lowercase();
        let disktype_enum = match disktype_lower.as_str() {
            "pcxt" => DiskType::PCXT,
            "pcat" => DiskType::PCAT,
            "ide" => DiskType::IDE,
            _ => DiskType::UNKNOWN,
        };

        let os_enum = OperatingSystem::from_str(os).map_err(|_| {
            std::io::Error::new(std::io::ErrorKind::InvalidInput, "Invalid operating system")
        })?;

        Ok(Disk {
            file: None,
            partition_table: PartitionTable::new(),
            metadata: Vhd::new(),
            disktype: disktype_enum,
            storage: BTreeMap::new(),
            os: os_enum,
            dirty: false,
        })
    }

    /// Attach an disk image file to the Disk struct for further use.
    pub fn attach(&mut self, filename: &str) -> Result<(), std::io::Error> {
        self.file = Some(
            OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .truncate(false)
                .open(filename)?,
        );
        match &mut self.file {
            Some(file) => match file.metadata() {
                Ok(metadata) => {
                    if metadata.len() > 512 {
                        self.load_metadata()?;
                        self.load_storage()?;
                    }
                    Ok(())
                }
                Err(_) => Err(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "Failed to attach the requested file.",
                )),
            },
            None => Ok(()),
        }
    }

    /// Calculate CHS geometry for disk based on type and size in Megabytes, rounded down to fit.
    pub fn set_geometry(&mut self, size: u32) -> Result<CHS, std::io::Error> {
        let max_spt = match self.disktype {
            DiskType::PCXT => 17,
            DiskType::PCAT => 17,
            DiskType::IDE => 63,
            DiskType::UNKNOWN => 63,
        };
        let max_heads = match self.disktype {
            DiskType::PCXT => 8,
            DiskType::PCAT => 15,
            DiskType::IDE => 16,
            DiskType::UNKNOWN => 16,
        };
        let max_size: u32 = 512 * max_spt * max_heads / 1024;
        if size <= max_size {
            let sector_count = (size * 1024 * 1024) / 512 - ((size / 512) % 512);
            let sectors_per_cylinder = max_heads * max_spt;
            let cylinders = sector_count / sectors_per_cylinder;
            let geometry = CHS::new(cylinders as u16, max_heads as u8, max_spt as u8)?;
            let exact_size = (cylinders * max_heads * max_spt * 512) + 512;
            if !self.dirty {
                self.dirty = true;
            }
            match &self.file {
                Some(file) => {
                    file.set_len(exact_size as u64)?;
                }
                None => {
                    return Err(std::io::Error::new(
                        std::io::ErrorKind::Other,
                        "Attach a file first. Current Disk has no backing storage.",
                    ))
                }
            };
            self.metadata.set_size(exact_size);
            self.metadata.set_geometry(geometry);
            if !self.dirty {
                self.dirty = true;
            }
            Ok(geometry)
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Cannot calculate geometry. Disk is too big for its type.",
            ))
        }
    }

    /// Get the disk's physical geometry
    pub fn get_geometry(&self) -> Result<CHS, std::io::Error> {
        Ok(*self.metadata.get_geometry())
    }

    /// Get the Operating System
    pub fn get_os(&self) -> OperatingSystem {
        self.os
    }

    /// Get the VHD metadata
    pub fn get_metadata(&self) -> &Vhd {
        &self.metadata
    }

    pub fn get_image_type(&self) -> &VhdDisktype {
        self.metadata.get_type()
    }
    
    /// Read a sector-sized block from the cache
    pub fn read_sector(&self, offset: u32) -> Result<[u8; 512], std::io::Error> {
        let sector_count = self.metadata.get_currentsize() / 512;
        if offset as u64 > sector_count {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Tried to read from beyond the disk's size.",
            ));
        }
        match &self.storage.get(&offset) {
            None => Ok([0u8; 512]),
            Some(sector) => Ok(**sector),
        }
    }

    /// Write a sector-sized block to the cache
    pub fn write_sector(&mut self, buffer: [u8; 512], offset: u32) -> Result<(), std::io::Error> {
        let sector_count = self.metadata.get_currentsize() / 512;
        if offset as u64 > sector_count {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Tried to write beyond the disk's size.",
            ));
        }
        self.storage.insert(offset, buffer);
        self.dirty = true;
        Ok(())
    }

    /// Write OS MBR boot code
    pub fn write_os_bootcode(&mut self) -> Result<(), std::io::Error> {
        let bytes = self.os.get_mbr_bootcode();
        let mut sector = self.read_sector(0)?;
        sector[0..446].copy_from_slice(&bytes);
        self.write_sector(sector, 0)?;
        Ok(())
    }

    /// Write MBR boot signature
    pub fn write_boot_signature(&mut self) -> Result<(), std::io::Error> {
        let mut sector = self.read_sector(0)?;
        sector[510] = 0x55;
        sector[511] = 0xaa;
        self.write_sector(sector, 0)?;
        Ok(())
    }

    pub fn write_partition_table(&mut self) -> Result<(), std::io::Error> {
        let mut sector = self.read_sector(0)?;
        let bytes = self.partition_table.as_bytes();
        sector[446..446 + 48].copy_from_slice(&bytes);
        self.write_sector(sector, 0)?;
        Ok(())
    }

    /// Load a Disk struct from a file
    pub fn load(path: &str) -> Result<Self, std::io::Error> {
        let mut disk = Disk::new("unknown", "unknown")?;
        disk.attach(path)?;
        disk.load_metadata()?;
        disk.load_partition_table()?;
        disk.load_storage()?;
        Ok(disk)
    }

    /// Loads VHD format metadata from the attached file.
    ///
    /// Attempts to load the VHD format metadata from the disk's attached file.
    /// If no file is attached to the disk, returns an `std::io::Error`.
    ///
    /// # Errors
    ///
    /// Returns an `std::io::Error` if:
    /// - No file is attached to the disk.
    /// - There's an error loading the VHD format metadata from the file.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::Disk;
    ///
    /// let mut disk = Disk::new();
    /// disk.attach_file("disk.vhd").unwrap();
    /// disk.load_metadata().unwrap();
    /// ```
    pub fn load_metadata(&mut self) -> Result<(), std::io::Error> {
        match &mut self.file {
            None => Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Attach a file first. The current disk has no backing storage.",
            )),
            Some(file) => {
                // Attempt to load VHD format metadata from the attached file
                match Vhd::load(file) {
                    Ok(metadata) => {
                        self.metadata = metadata;
                        Ok(())
                    }
                    Err(err) => Err(std::io::Error::new(
                        std::io::ErrorKind::Other,
                        format!("Failed to load VHD format metadata: {}", err),
                    )),
                }
            }
        }
    }
    /// Loads the partition table from the disk's attached file.
    ///
    /// Reads the first sector of the disk's attached file to load the partition table.
    /// If no file is attached to the disk, returns an `std::io::Error`.
    ///
    /// # Errors
    ///
    /// Returns an `std::io::Error` if:
    /// - No file is attached to the disk.
    /// - There's an error loading the partition table from the disk.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::Disk;
    ///
    /// let mut disk = Disk::new();
    /// disk.attach_file("disk.img").unwrap();
    /// disk.load_partition_table().unwrap();
    /// ```
    pub fn load_partition_table(&mut self) -> Result<(), std::io::Error> {
        // Ensure there's a file attached to the disk
        if self.file.is_some() {
            // Borrow self mutably to load the partition table
            let mut partition_table = PartitionTable::new(); // Or any other way to initialize the partition table
            partition_table.load(self)?;

            // Assign the loaded partition table to self.partition_table
            self.partition_table = partition_table;

            Ok(())
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Attach a file first. Current Disk has no backing storage.",
            ))
        }
    }

    /// Load bytes from file into storage.
    fn load_storage(&mut self) -> Result<(), std::io::Error> {
        match &mut self.file {
            Some(file) => {
                let mut buffer = [0u8; 512]; // Storage for a single sector
                let mut sector_number: u32 = 0; // Start value for the sector number.
                file.seek(SeekFrom::Start(0))?;
                loop {
                    let bytes_read = file.read(&mut buffer)?;
                    if bytes_read == 0 {
                        break;
                    }
                    if !buffer.iter().all(|&byte| byte == 0) {
                        let mut cloned_buffer = [0u8; 512];
                        cloned_buffer.copy_from_slice(&buffer);
                        self.storage.insert(sector_number, cloned_buffer);
                    }
                    sector_number += 1;
                }
                if let Some((&max_key, _)) = self.storage.iter().next_back() {
                    // Remove the entry with the maximum key
                    if self.storage.remove_entry(&max_key).is_some() {
                        return Ok(());
                    } else {
                        return Err(std::io::Error::new(
                            std::io::ErrorKind::Other,
                            "Something strange with this file. Are you sure it is at least 512 bytes in size?",
                        ));
                    }
                }
                Ok(())
            }
            None => Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Attach a file first. Current Disk has no backing storage.",
            )),
        }
    }

    /// Commit the Disk struct to permanent storage.
    pub fn commit_storage(&mut self) -> Result<(), std::io::Error> {
        if self.dirty {
            match &mut self.file {
                Some(file) => {
                    // Write sectors from the cache
                    for (offset, sector) in &self.storage {
                        file.seek(SeekFrom::Start(*offset as u64 * 512))?;
                        file.write_all(sector)?;
                    }
                    // Write the metadata to the last 512 bytes
                    file.seek(SeekFrom::End(-512))?;
                    file.write_all(&self.metadata.as_bytes())?;
                    self.dirty = false;
                    Ok(())
                }
                None => Err(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "Attach a file first. Current Disk has no backing storage.",
                )),
            }
        }
        // Cache is clean, treat the commit request as a NOP.
        else {
            Ok(())
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use tempfile::NamedTempFile;

    #[test]
    fn test_set_geometry_pcxt_small_disk_attached() {
        let mut disk = Disk::new("pcxt", "ibmdos200").unwrap();
        let file = NamedTempFile::new().expect("Failed to create tempfile");
        let file_path = file.path().to_str().unwrap(); // Get the path as a string
        disk.attach(file_path).unwrap(); // Pass the path as a string
        let result = disk.set_geometry(10); // 10 MB
        assert!(result.is_ok());
        let geometry = result.unwrap();
        assert_eq!(geometry.cylinder, 150); // Assuming 8 heads and 17 sectors per track
    }

    #[test]
    fn test_new_pcxt() {
        let disk = Disk::new("pcxt", "ibmdos200");
        assert!(disk.is_ok());
        let disk = disk.unwrap();
        assert_eq!(disk.disktype, DiskType::PCXT);
        assert_eq!(disk.os, OperatingSystem::IBMDOS200);
    }

    #[test]
    fn test_new_pcat() {
        let disk = Disk::new("pcat", "ibmdos200");
        assert!(disk.is_ok());
        let disk = disk.unwrap();
        assert_eq!(disk.disktype, DiskType::PCAT);
        assert_eq!(disk.os, OperatingSystem::IBMDOS200);
    }

    #[test]
    fn test_new_ide() {
        let disk = Disk::new("ide", "ibmdos200");
        assert!(disk.is_ok());
        let disk = disk.unwrap();
        assert_eq!(disk.disktype, DiskType::IDE);
        assert_eq!(disk.os, OperatingSystem::IBMDOS200);
    }

    #[test]
    fn test_new_unknown_disktype() {
        let disk = Disk::new("unknown", "ibmdos200");
        assert!(disk.is_ok());
        let disk = disk.unwrap();
        assert_eq!(disk.disktype, DiskType::UNKNOWN);
        assert_eq!(disk.os, OperatingSystem::IBMDOS200);
    }

    #[test]
    fn test_new_invalid_os() {
        let disk = Disk::new("pcxt", "InvalidOS");
        assert!(disk.is_ok());
        assert_eq!(disk.unwrap().disktype, DiskType::PCXT);
    }
}
