use core::fmt;

/// Custom type for Cylinder/Head/Sector geometry
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct CHS {
    pub cylinder: u16,
    pub head: u8,
    pub sector: u8,
}

impl Default for CHS {
    fn default() -> Self {
        CHS::empty()
    }
}

impl CHS {
    /// Instantiate a new CHS struct.
    ///
    /// # Arguments
    ///
    /// * `cylinder` - The cylinder number.
    /// * `head` - The head number.
    /// * `sector` - The sector number.
    ///
    /// # Returns
    ///
    /// A Result containing the newly created CHS struct if successful, or an error if
    /// the provided parameters are out of bounds.
    pub fn new(cylinder: u16, head: u8, sector: u8) -> Result<CHS, std::io::Error> {
        if cylinder > 1024 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Too many cylinders.",
            ));
        }
        if sector > 63 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Too many sectors.",
            ));
        }
        Ok(CHS {
            cylinder,
            head,
            sector,
        })
    }

    /// Creates an empty CHS (Cylinder-Head-Sector) struct with all fields set to zero.
    pub fn empty() -> CHS {
        CHS {
            cylinder: 0,
            head: 0,
            sector: 0,
        }
    }

    /// Converts the CHS (Cylinder-Head-Sector) to the 4 bytes required for the VHD disk format footer field.
    /// This conversion is different and simpler than the MBR partition entry calculations.
    pub fn as_vhd_bytes(&self) -> [u8; 4] {
        let cylinder_bytes = self.cylinder.to_be_bytes();
        [cylinder_bytes[0], cylinder_bytes[1], self.head, self.sector]
    }

    /// Convert the 4 bytes from the VHD disk format to CHS values.
    /// This is the inverse operation of `as_vhd_bytes`.
    pub fn from_vhd_bytes(bytes: [u8; 4]) -> CHS {
        let cylinder = u16::from_be_bytes([bytes[0], bytes[1]]);
        let head = bytes[2];
        let sector = bytes[3];

        CHS {
            cylinder,
            head,
            sector,
        }
    }

    /// Calculate triplet of CHS bytes for use in partition tables
    /// <https://thestarman.pcministry.com/asm/mbr/PartTables.htm#mbr>
    pub fn as_mbr_bytes(&self) -> [u8; 3] {
        let sector_bits = self.sector & 0b00111111;
        let cylinder_byte = (self.cylinder & 0b0000000011111111) as u8;
        let cylinder_overflow_bits = self.cylinder & 0b0000001100000000;
        let sector_byte = sector_bits | (cylinder_overflow_bits >> 2) as u8;
        [self.head, sector_byte, cylinder_byte]
    }

    /// Turn the encoded CHS-value from the bytes in an MBR to a CHS-tuple
    /// The order of the bytes is identical to the way they are encoded on-disk
    /// on an old MBR disk. So generally: Heads, Sectors, Cylinders in that order.
    pub fn from_bytes(bytes: [u8; 3]) -> CHS {
        let head = bytes[0];
        let sector_byte = bytes[1];
        let cylinder = (((sector_byte & 0b11000000) as u16) << 2) | (bytes[2] as u16);
        let sector = sector_byte & 0b00111111;
        CHS::new(cylinder, head, sector).unwrap() // [TODO] Lose the unwrap at some point
    }

    /// Provide a physical disk geometry to convert the current CHS into an LBA within this geometry.
    pub fn to_lba(&self, geometry: &CHS) -> Result<u32, std::io::Error> {
        if geometry.cylinder < self.cylinder
            || geometry.head < self.head
            || geometry.sector < self.sector
        {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Address does not fit inside drive geometry.",
            ));
        }
        let cylinder = u32::from(self.cylinder);
        let head = u32::from(self.head);
        let sector = u32::from(self.sector);

        let total_heads = u32::from(geometry.head);
        let total_sectors_per_track = u32::from(geometry.sector);

        let lba = (cylinder * (total_heads * total_sectors_per_track))
            + (head * total_sectors_per_track)
            + (sector - 1);

        Ok(lba)
    }

    /// Instantiate a CHS address from an LBA (Logical Block Address) and an existing disk geometry represented by a CHS (Cylinder-Head-Sector) address.
    ///
    /// # Arguments
    ///
    /// * `geometry` - A reference to the CHS address representing the disk geometry.
    /// * `lba` - The Logical Block Address (LBA) to convert into a CHS address.
    ///
    /// # Returns
    ///
    /// A Result containing the CHS address corresponding to the provided LBA if successful, or an `std::io::Error` if the conversion fails due to the LBA being out of bounds for the given disk geometry.
    ///
    /// # Errors
    ///
    /// Returns an `std::io::Error` if the provided LBA results in a CHS address that exceeds the bounds specified by the disk geometry.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::CHS;
    ///
    /// // Define disk geometry
    /// let geometry = CHS {
    ///     cylinder: 20,
    ///     head: 2,
    ///     sector: 10,
    /// };
    ///
    /// // Convert LBA to CHS address
    /// let lba = 201;
    /// match CHS::from_lba(&geometry, lba) {
    ///     Ok(chs_address) => {
    ///         println!("CHS Address: {:?}", chs_address);
    ///     }
    ///     Err(err) => {
    ///         eprintln!("Error: {}", err);
    ///     }
    /// }
    /// ```
    ///
    /// This example converts an LBA (201) to a CHS address using the given disk geometry and prints the resulting CHS address.
    ///
    /// # Note
    ///
    /// This method assumes that the LBA, cylinder, head, and sector values are within their valid ranges as defined by the CHS addressing scheme.
    pub fn from_lba(geometry: &CHS, lba: u32) -> Result<CHS, std::io::Error> {
        let mut chs = CHS::empty();
        let sectors_per_track = u32::from(geometry.sector);
        let heads_per_cylinder = u32::from(geometry.head);

        // Check if sectors_per_track or heads_per_cylinder is zero
        if sectors_per_track == 0 || heads_per_cylinder == 0 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Invalid geometry.",
            ));
        }

        let total_sectors = u32::from(geometry.cylinder) * sectors_per_track * heads_per_cylinder;
        if lba >= total_sectors {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "LBA out of bounds for geometry.",
            ));
        }
        // Calculate cylinder, head, and sector values from LBA
        chs.cylinder = u16::try_from(lba / (heads_per_cylinder * sectors_per_track))
            .map_err(|_| std::io::Error::new(std::io::ErrorKind::Other, "Too many cylinders."))?;
        let temp = lba % (heads_per_cylinder * sectors_per_track);
        chs.head = u8::try_from(temp / sectors_per_track)
            .map_err(|_| std::io::Error::new(std::io::ErrorKind::Other, "Too many heads."))?;
        chs.sector = u8::try_from(temp % sectors_per_track + 1)
            .map_err(|_| std::io::Error::new(std::io::ErrorKind::Other, "Too many sectors."))?;

        // Check bounds against geometry
        if geometry.cylinder < chs.cylinder {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Cylinder out of bounds for geometry.",
            ));
        }
        if geometry.head < chs.head {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Head out of bounds for geometry.",
            ));
        }
        if geometry.sector < chs.sector {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Sector out of bounds for geometry.",
            ));
        }
        Ok(chs)
    }

    /// Return the number of sectors that would fit in this geometry
    pub fn get_sector_count(&self) -> u32 {
        self.cylinder as u32 * self.head as u32 * self.sector as u32
    }
}

impl fmt::Display for CHS {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Cylinder: {} Head: {} Sector: {}",
            self.cylinder, self.head, self.sector
        )
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_new_chs_within_bounds() {
        // Test creating a CHS struct with valid parameters.
        let result = CHS::new(512, 4, 32);
        assert!(result.is_ok());
        let chs = result.unwrap();
        assert_eq!(chs.cylinder, 512);
        assert_eq!(chs.head, 4);
        assert_eq!(chs.sector, 32);
    }

    #[test]
    fn test_new_chs_cylinder_out_of_bounds() {
        // Test creating a CHS struct with a cylinder number out of bounds.
        let result = CHS::new(1100, 4, 32);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_new_chs_sector_out_of_bounds() {
        // Test creating a CHS struct with a sector number out of bounds.
        let result = CHS::new(512, 4, 64);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_empty_chs() {
        // Create an empty CHS struct.
        let empty_chs = CHS::empty();

        // Assert that all fields are set to zero.
        assert_eq!(empty_chs.cylinder, 0);
        assert_eq!(empty_chs.head, 0);
        assert_eq!(empty_chs.sector, 0);
    }

    #[test]
    fn test_as_vhd_bytes_zero_values() {
        // Test with all fields set to zero.
        let chs = CHS {
            cylinder: 0,
            head: 0,
            sector: 0,
        };
        let vhd_bytes = chs.as_vhd_bytes();
        assert_eq!(vhd_bytes, [0, 0, 0, 0]);
    }

    #[test]
    fn test_as_vhd_bytes_max_values() {
        // Test with maximum values for each field.
        let chs = CHS {
            cylinder: 65535,
            head: 255,
            sector: 255,
        };
        let vhd_bytes = chs.as_vhd_bytes();
        assert_eq!(vhd_bytes, [255, 255, 255, 255]);
    }

    #[test]
    fn test_as_vhd_bytes_mid_range_values() {
        // Test with mid-range values for each field.
        let chs = CHS {
            cylinder: 512,
            head: 4,
            sector: 32,
        };
        let vhd_bytes = chs.as_vhd_bytes();
        assert_eq!(vhd_bytes, [2, 0, 4, 32]);
    }

    #[test]
    fn test_as_vhd_bytes_mixed_values() {
        // Test with mixed values.
        let chs = CHS {
            cylinder: 1024,
            head: 8,
            sector: 63,
        };
        let vhd_bytes = chs.as_vhd_bytes();
        assert_eq!(vhd_bytes, [4, 0, 8, 63]);
    }

    #[test]
    fn test_as_mbr_bytes_zero_values() {
        // Test with all fields set to zero.
        let chs = CHS {
            cylinder: 0,
            head: 0,
            sector: 0,
        };
        let mbr_bytes = chs.as_mbr_bytes();
        assert_eq!(mbr_bytes, [0, 0, 0]);
    }

    #[test]
    fn test_as_mbr_bytes_max_values() {
        // Test with maximum values for each field.
        let chs = CHS {
            cylinder: 65535,
            head: 255,
            sector: 63,
        };
        let mbr_bytes = chs.as_mbr_bytes();
        assert_eq!(mbr_bytes, [0xFF, 0xFF, 0xFF]);
    }

    #[test]
    fn test_as_mbr_bytes_mid_range_values() {
        // Test with mid-range values for each field.
        let chs = CHS {
            cylinder: 512,
            head: 4,
            sector: 32,
        };
        let mbr_bytes = chs.as_mbr_bytes();
        assert_eq!(mbr_bytes, [4, 160, 0]);
    }

    #[test]
    fn test_as_mbr_bytes_mixed_values() {
        // Test with mixed values.
        let chs = CHS {
            cylinder: 1024,
            head: 8,
            sector: 63,
        };
        let mbr_bytes = chs.as_mbr_bytes();
        assert_eq!(mbr_bytes, [8, 63, 0]);
    }

    #[test]
    fn test_to_lba_valid_conversion() {
        // Test with valid CHS and disk geometry.
        let chs = CHS {
            cylinder: 10,
            head: 1,
            sector: 1,
        };
        let disk_geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let result = chs.to_lba(&disk_geometry);
        assert_eq!(result.unwrap(), 210);
    }

    #[test]
    fn test_to_lba_chs_exceeds_disk_geometry() {
        // Test when CHS parameters exceed disk geometry.
        let chs = CHS {
            cylinder: 30,
            head: 3,
            sector: 15,
        };
        let disk_geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let result = chs.to_lba(&disk_geometry);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_to_lba_invalid_geometry() {
        // Test when disk geometry has zero heads.
        let chs = CHS {
            cylinder: 10,
            head: 1,
            sector: 1,
        };
        let disk_geometry = CHS {
            cylinder: 20,
            head: 0,
            sector: 10,
        };
        let result = chs.to_lba(&disk_geometry);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_from_lba_valid_conversion() {
        // Test with valid LBA and disk geometry.
        let geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let lba = 201;
        let result = CHS::from_lba(&geometry, lba).unwrap();
        assert_eq!(
            result,
            CHS {
                cylinder: 10,
                head: 0,
                sector: 2
            }
        );
    }

    #[test]
    fn test_mbr_bytes_roundtrip() {
        let base = CHS::new(523, 7, 15).unwrap();
        let bytes = base.as_mbr_bytes();
        let roundtripped = CHS::from_bytes(bytes);
        assert_eq!(base, roundtripped);
    }

    #[test]
    fn test_from_lba_invalid_cylinder() {
        // Test when LBA results in a cylinder out of bounds.
        let geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let lba = 10000;
        let result = CHS::from_lba(&geometry, lba);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_from_lba_invalid_head() {
        // Test when LBA results in a head out of bounds.
        let geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let lba = 1000;
        let result = CHS::from_lba(&geometry, lba);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_from_lba_invalid_sector() {
        // Test when LBA results in a sector out of bounds.
        let geometry = CHS {
            cylinder: 20,
            head: 2,
            sector: 10,
        };
        let lba = 3000;
        let result = CHS::from_lba(&geometry, lba);
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().kind(), std::io::ErrorKind::Other);
    }
}
