use super::{chs::CHS, Disk};
use crate::fat::Fat;
use std::fmt;

#[derive(Clone, Debug)]
pub struct Partition {
    status: PartitionStatus,
    start_sector: CHS,
    partition_type: PartitionType,
    end_sector: CHS,
    start_sector_lba: u32,
    sector_count: u32,
    filesystem: Option<Fat>,
}

impl fmt::Display for Partition {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.partition_type {
            PartitionType::Empty => {
                writeln!(f, "Partition type   : Empty")
            }
            _ => {
                write!(
                    f,
                    "Partition type   : {}\n\
             Start address    : {}\n\
             Size             : {} bytes\n\
             Sector count     : {}\n\
             Start sector LBA : {}\n\
             End address      : {}\n",
                    self.partition_type,
                    self.start_sector,
                    self.sector_count * 512,
                    self.sector_count,
                    self.start_sector_lba,
                    self.end_sector,
                )
            }
        }
    }
}

impl Default for Partition {
    fn default() -> Self {
        Self::new()
    }
}

impl Partition {
    /// Creates a new partition with placeholder values indicating unknown.
    ///
    /// This method creates a new `Partition` instance with placeholder values for its fields,
    /// indicating that their actual values are unknown or uninitialized. These placeholder
    /// values can be updated later as needed.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Partition;
    ///
    /// let partition = Partition::new();
    /// ```
    pub fn new() -> Self {
        Partition {
            status: PartitionStatus::Inactive,
            start_sector: CHS::default(), // Placeholder for unknown start sector
            partition_type: PartitionType::Empty,
            end_sector: CHS::default(), // Placeholder for unknown end sector
            start_sector_lba: 0,        // Placeholder for unknown start sector LBA
            sector_count: 0,            // Placeholder for unknown sector count
            filesystem: None,
        }
    }

    /// Reads a sector from the disk.
    ///
    /// # Arguments
    ///
    /// * `disk` - A reference to the `Disk` object representing the disk from which to read.
    /// * `sector_address` - The address of the sector to read, specified as a 32-bit unsigned integer.
    ///
    /// # Returns
    ///
    /// A result containing a fixed-size array of 512 bytes representing the data read from the sector,
    /// or an `std::io::Error` if an I/O error occurs during the read operation.
    pub fn read_sector(
        &self,
        disk: &Disk,
        sector_address: u32,
    ) -> Result<[u8; 512], std::io::Error> {
        let disk_sector = sector_address + self.start_sector_lba;
        let data = disk.read_sector(disk_sector)?;
        Ok(data)
    }

    /// Writes data to a sector on the disk.
    ///
    /// # Arguments
    ///
    /// * `disk` - A mutable reference to the `Disk` object representing the disk to write to.
    /// * `sector_address` - The address of the sector to write to, specified as a 32-bit unsigned integer.
    /// * `data` - A fixed-size array of 512 bytes representing the data to write to the sector.
    ///
    /// # Returns
    ///
    /// A result indicating success (`Ok(())`) if the write operation is successful,
    /// or an `std::io::Error` if an I/O error occurs during the write operation.
    pub fn write_sector(
        &self,
        disk: &mut Disk,
        sector_address: u32,
        data: [u8; 512],
    ) -> Result<(), std::io::Error> {
        let disk_sector = sector_address + self.start_sector_lba;
        disk.write_sector(data, disk_sector)?;
        Ok(())
    }

    /// Creates a single partition that's active and as large as possible on the given disk.
    ///
    /// # Arguments
    ///
    /// * `disk` - A reference to the disk on which the partition will be created.
    /// * `partition_type` - Optional partition type. Defaults to `Fat12MBR` if not specified.
    ///
    /// # Errors
    ///
    /// Returns an error if there is an issue with disk operations or CHS conversions.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::{Disk, Partition, PartitionType};
    ///
    /// let disk = Disk::new();
    /// let partition = Partition::fill_disk(&disk, Some(PartitionType::Fat32));
    /// ```
    pub fn fill_disk(
        disk: &Disk,
        partition_type: Option<PartitionType>,
    ) -> Result<Self, std::io::Error> {
        // Get disk geometry
        let geometry = disk.get_geometry()?;

        // Get total sector count
        let sector_count = geometry.get_sector_count();

        // Ensure the disk has non-zero size
        if sector_count == 0 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Disk size is zero",
            ));
        }

        Ok(Partition {
            status: PartitionStatus::Active,
            start_sector: CHS::new(0, 0, 2)?,
            partition_type: partition_type.unwrap_or(PartitionType::Fat12MBR),
            end_sector: CHS::from_lba(&geometry, sector_count - 3)?,
            start_sector_lba: CHS::new(0, 0, 2)?.to_lba(&geometry)?,
            sector_count,
            filesystem: None,
        })
    }

    /// Set the starting sector by LBA. This method updates the start_sector CHS field and recalculates the sector_count field as needed to match.
    ///
    /// # Arguments
    ///
    /// * `disk` - A reference to the disk on which the partition resides.
    /// * `lba` - The Logical Block Address (LBA) indicating the starting sector of the partition.
    ///
    /// # Errors
    ///
    /// Returns an error if there is an issue with disk operations or if the LBA calculation results in invalid values.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::{Disk, Partition};
    ///
    /// let mut disk = Disk::new("pcxt", "ibmdos200").unwrap();
    /// let mut partition = Partition::new();
    ///
    /// partition.set_start_lba(&disk, 5).unwrap();
    /// ```
    ///
    /// # Remarks
    ///
    /// This method updates the start_sector CHS field and recalculates the sector_count field based on the provided LBA value and the disk's geometry.
    pub fn set_start_lba(&mut self, disk: &Disk, lba: u32) -> Result<(), std::io::Error> {
        let geometry: CHS = disk.get_geometry()?;
        let chs_address = CHS::from_lba(&geometry, lba)?;
        self.start_sector = chs_address;
        self.start_sector_lba = lba;

        // Calculate sector count only if end_sector is valid
        if self.end_sector.cylinder > 0 && self.end_sector.head > 0 && self.end_sector.sector > 0 {
            let end_lba = self.end_sector.to_lba(&geometry)?;
            self.sector_count = end_lba - lba;
        }
        Ok(())
    }

    /// Sets the last sector by LBA. This effectively updates the sector count and end_sector CHS fields.
    ///
    /// # Arguments
    ///
    /// * `disk` - A reference to the disk on which the partition resides.
    /// * `lba` - The Logical Block Address (LBA) indicating the ending sector of the partition.
    ///
    /// # Errors
    ///
    /// Returns an error if the requested LBA exceeds the disk's total sector count or if the requested LBA is before the partition's first sector.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::{Disk, Partition};
    ///
    /// let mut disk = Disk::new("pcxt", "ibmdos200").unwrap();
    /// let mut partition = Partition::new();
    ///
    /// partition.set_start_lba(&disk, 5).unwrap();
    /// partition.set_end_lba(&disk, 10).unwrap();
    /// ```
    pub fn set_end_lba(&mut self, disk: &Disk, lba: u32) -> Result<(), std::io::Error> {
        let geometry = disk.get_geometry()?;
        let disk_last_sector = geometry.to_lba(&geometry)?;

        if lba > disk_last_sector {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Requested end LBA is higher than the disk's total sector count.",
            ));
        }

        if lba <= self.start_sector_lba {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Partition's last sector cannot be located before its first sector.",
            ));
        }

        self.end_sector = CHS::from_lba(&geometry, lba)?;
        self.sector_count = lba - self.start_sector_lba;
        Ok(())
    }

    /// Sets the status field according to an incoming string: "active" or "inactive". Any other string will yield a PartitionStatus::Invalid.
    ///
    /// # Arguments
    ///
    /// * `status` - A string indicating the desired status ("active" or "inactive").
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Partition;
    ///
    /// let mut partition = Partition::new();
    ///
    /// partition.set_status("active");
    /// assert_eq!(partition.status, PartitionStatus::Active);
    ///
    /// partition.set_status("inactive");
    /// assert_eq!(partition.status, PartitionStatus::Inactive);
    ///
    /// partition.set_status("invalid");
    /// assert_eq!(partition.status, PartitionStatus::Invalid);
    /// ```
    pub fn set_status(&mut self, status: &str) {
        self.status = PartitionStatus::from_str(status);
    }

    /// Returns the value of the status field from the Partition object.
    pub fn get_status(&self) -> PartitionStatus {
        self.status
    }

    /// Sets the partition type by hex number.
    ///
    /// The partition type is represented by a hex number according to the MBR partition table specification.
    /// For more information, see: [Wikipedia - Partition type](https://en.wikipedia.org/wiki/Partition_type).
    ///
    /// # Arguments
    ///
    /// * `byte` - The hex number representing the partition type.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::{Partition, PartitionType};
    ///
    /// let mut partition = Partition::new();
    ///
    /// partition.set_partition_type(0x01);
    /// assert_eq!(partition.partition_type, PartitionType::Fat12MBR);
    ///
    /// partition.set_partition_type(0x07);
    /// assert_eq!(partition.partition_type, PartitionType::NTFS);
    /// ```
    pub fn set_partition_type(&mut self, byte: u8) {
        self.partition_type = PartitionType::from_byte(byte);
    }

    /// Converts the struct to the on-disk partition entry representation.
    ///
    /// # Returns
    ///
    /// An array representing the partition entry in the on-disk format.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Partition;
    ///
    /// let partition = Partition::new();
    /// let partition_entry = partition.as_partition_entry();
    /// ```
    pub fn as_partition_entry(&self) -> [u8; 16] {
        let mut entry = [0u8; 16];

        // Set status byte
        entry[0] = self.status.as_byte();

        // Copy start sector CHS bytes
        entry[1..4].copy_from_slice(&self.start_sector.as_mbr_bytes());

        // Set partition type byte
        entry[4] = self.partition_type.as_byte();

        // Copy end sector CHS bytes
        entry[5..8].copy_from_slice(&self.end_sector.as_mbr_bytes());

        // Copy start LBA bytes (little-endian)
        entry[9..13].copy_from_slice(&self.start_sector_lba.to_le_bytes());

        // Copy sector count bytes (little-endian)
        entry[12..16].copy_from_slice(&self.sector_count.to_le_bytes());

        entry
    }

    /// Get a reference to the filesystem associated with the partition, if any.
    ///
    /// # Returns
    ///
    /// A reference to the filesystem wrapped in an `Option`.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Partition;
    ///
    /// let partition = Partition::new();
    /// let filesystem = partition.get_filesystem();
    /// ```
    pub fn get_filesystem(&self) -> Option<&Fat> {
        self.filesystem.as_ref()
    }

    pub fn get_start_sector(&self) -> &CHS {
        &self.start_sector
    }

    pub fn get_partition_type(&self) -> &PartitionType {
        &self.partition_type
    }

    pub fn get_sector_count(&self) -> u32 {
        self.sector_count
    }

    /// Converts the struct into on-disk partition entry bytes.
    ///
    /// # Returns
    ///
    /// An array representing the partition entry in the on-disk format.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Partition;
    ///
    /// let partition = Partition::new();
    /// let partition_entry = partition.as_bytes();
    /// ```
    pub fn as_bytes(&self) -> [u8; 16] {
        let mut entry = [0u8; 16];
        entry[0] = self.status.as_byte();
        entry[1..4].copy_from_slice(&self.start_sector.as_mbr_bytes());
        entry[4] = self.partition_type.as_byte();
        entry[5..8].copy_from_slice(&self.end_sector.as_mbr_bytes());
        entry[8..12].copy_from_slice(&self.start_sector_lba.to_le_bytes());
        entry[12..16].copy_from_slice(&self.sector_count.to_le_bytes());
        entry
    }

    /// Converts the on-disk partition entry into a struct.
    ///
    /// # Arguments
    ///
    /// * `bytes` - A reference to the array containing the partition entry bytes.
    ///
    /// # Returns
    ///
    /// A `Result` containing either a `Partition` struct representing the partition entry
    /// or an `std::io::Error` if there was an error during conversion.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::{Partition, PartitionStatus, CHS, PartitionType};
    ///
    /// let bytes: [u8; 16] = [
    ///     0x80,  // Active status byte
    ///     1, 2, 3,  // Start sector CHS bytes
    ///     0x0C,  // FAT32 partition type
    ///     4, 5, 6,  // End sector CHS bytes
    ///     10, 0, 0, 0,  // Start sector LBA address
    ///     20, 0, 0, 0,  // Sector count
    /// ];
    ///
    /// let partition = Partition::from_bytes(&bytes).unwrap();
    /// assert_eq!(partition.status, PartitionStatus::Active);
    /// assert_eq!(partition.start_sector, CHS::new(1, 2, 3).unwrap());
    /// assert_eq!(partition.partition_type, PartitionType::Fat32);
    /// assert_eq!(partition.end_sector, CHS::new(4, 5, 6).unwrap());
    /// assert_eq!(partition.start_sector_lba, 10);
    /// assert_eq!(partition.sector_count, 20);
    /// ```
    pub fn from_bytes(bytes: &[u8; 16]) -> Result<Partition, std::io::Error> {
        let status = PartitionStatus::from_byte(bytes[0]);
        let mut start_sector_bytes = [0u8; 3];
        start_sector_bytes.copy_from_slice(&bytes[1..4]);
        let start_sector = CHS::from_bytes(start_sector_bytes);

        let partition_type = PartitionType::from_byte(bytes[4]);

        let mut end_sector_bytes = [0u8; 3];
        end_sector_bytes.copy_from_slice(&bytes[5..8]);
        
        let end_sector = CHS::from_bytes(end_sector_bytes);

        let start_sector_lba = u32::from_le_bytes(bytes[8..12].try_into().map_err(|_| {
            std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Failed to read LBA address for the start sector.",
            )
        })?);

        let sector_count = u32::from_le_bytes(bytes[12..16].try_into().map_err(|_| {
            std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Failed to read the sector count field.",
            )
        })?);

        Ok(Partition {
            status,
            start_sector,
            partition_type,
            end_sector,
            start_sector_lba,
            sector_count,
            filesystem: None,
        })
    }
}

#[derive(PartialEq, Clone, Copy, Debug)]
pub enum PartitionStatus {
    Active,
    Inactive,
    Invalid,
}

impl PartitionStatus {
    pub fn from_str(string: &str) -> Self {
        match string {
            "active" => Self::Active,
            "inactive" => Self::Inactive,
            _ => Self::Invalid,
        }
    }

    pub fn from_byte(byte: u8) -> Self {
        match byte {
            0x00 => Self::Inactive,
            0x80 => Self::Active,
            _ => Self::Invalid,
        }
    }
    pub fn as_byte(&self) -> u8 {
        match self {
            Self::Active => 0x80,
            Self::Inactive => 0x00,
            _ => 0x01,
        }
    }
}

impl Default for PartitionStatus {
    fn default() -> Self {
        Self::Inactive
    }
}

#[derive(PartialEq, Clone, Copy, Debug)]
pub enum PartitionType {
    Empty,
    Fat12MBR,
    Fat16MBR,
    ExtendedCHS,
    Fat16B,
}

impl fmt::Display for PartitionType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let display_value = match self {
            Self::Empty => "Empty",
            Self::Fat12MBR => "FAT12 MBR formatted hard disk",
            Self::Fat16MBR => "FAT16 MS-DOS 3.3x compatible.",
            Self::Fat16B => "FAT16B 'BIGDOS'",
            Self::ExtendedCHS => "Extended partition",
        };
        // Use \n for newlines to format the output across multiple lines
        write!(f, "{}", display_value)
    }
}

impl PartitionType {
    pub fn from_byte(byte: u8) -> PartitionType {
        match byte {
            0x01 => Self::Fat12MBR,
            0x04 => Self::Fat16MBR,
            0x05 => Self::ExtendedCHS,
            0x06 => Self::Fat16B,
            _ => Self::Empty,
        }
    }

    pub fn as_byte(&self) -> u8 {
        match self {
            Self::Empty => 0x00,
            Self::Fat12MBR => 0x01,
            Self::Fat16MBR => 0x04,
            Self::ExtendedCHS => 0x05,
            Self::Fat16B => 0x06,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use tempfile::Builder;

    #[test]
    fn test_new() {
        // Create a new partition
        let partition = Partition::new();

        // Check that default values are set
        assert_eq!(partition.status, PartitionStatus::Inactive);
        assert_eq!(partition.start_sector, CHS::default());
        assert_eq!(partition.partition_type, PartitionType::Empty);
        assert_eq!(partition.end_sector, CHS::default());
        assert_eq!(partition.start_sector_lba, 0);
        assert_eq!(partition.sector_count, 0);
        assert_eq!(partition.filesystem, None);
    }

    #[test]
    fn test_fill_disk_with_valid_disk() {
        // Create a temporary file for the disk
        let disk_file = Builder::new().suffix(".img").tempfile().unwrap();
        let disk_path = disk_file.path().to_str().unwrap();

        // Create a disk and set its geometry
        let mut disk = Disk::new("pcxt", "ibmdos200").unwrap();
        disk.attach(disk_path).unwrap();
        disk.set_geometry(10).unwrap(); // 10 MB disk for testing

        // Call the fill_disk method
        let partition = Partition::fill_disk(&disk, None).unwrap();

        // Ensure that start sector is (0, 0, 2)
        assert_eq!(partition.start_sector, CHS::new(0, 0, 2).unwrap());
        // Ensure that end sector is calculated correctly
        assert_eq!(
            partition.end_sector,
            CHS::from_lba(&disk.get_geometry().unwrap(), partition.sector_count - 3).unwrap()
        );
        // Ensure that sector count is correct
        assert_eq!(
            partition.sector_count,
            disk.get_geometry().unwrap().get_sector_count()
        );
        // Ensure that filesystem is None
        assert_eq!(partition.filesystem, None);
    }

    #[test]
    fn test_fill_disk_with_specified_partition_type() {
        // Create a temporary file for the disk
        let disk_file = Builder::new().suffix(".img").tempfile().unwrap();
        let disk_path = disk_file.path().to_str().unwrap();

        // Create a disk and set its geometry
        let mut disk = Disk::new("pcxt", "ibmdos200").unwrap();
        disk.attach(disk_path).unwrap();
        disk.set_geometry(10).unwrap(); // 10 MB disk for testing

        // Call the fill_disk method with a specified partition type
        let partition = Partition::fill_disk(&disk, Some(PartitionType::Fat12MBR)).unwrap();

        // Ensure that the partition type is set correctly
        assert_eq!(partition.partition_type, PartitionType::Fat12MBR);
    }

    #[test]
    fn test_fill_disk_with_empty_disk() {
        // Create a disk object without attaching it
        let disk = Disk::new("pcxt", "ibmdos200").unwrap();

        // Call the fill_disk method with an empty disk
        let result = Partition::fill_disk(&disk, None);

        // Ensure that an error is returned
        assert!(result.is_err());
        // Ensure that the error type is Other
        assert_eq!(result.err().unwrap().kind(), std::io::ErrorKind::Other);
    }
    #[test]
    fn test_fill_disk_with_invalid_disk() {
        // Create a disk object without attaching it
        let disk = Disk::new("pcxt", "ibmdos200").unwrap();

        // Call the fill_disk method with an invalid disk
        let result = Partition::fill_disk(&disk, None);

        // Ensure that an error is returned
        assert!(result.is_err());
        // Ensure that the error type is Other
        assert_eq!(result.err().unwrap().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_set_start_lba_with_invalid_end_sector() {
        let mut partition = Partition {
            start_sector: CHS::new(0, 0, 2).unwrap(),
            start_sector_lba: 2,
            end_sector: CHS::default(),
            sector_count: 0,
            filesystem: None,
            partition_type: PartitionType::Fat12MBR,
            status: PartitionStatus::Inactive,
        };

        let disk = Disk::new("pcxt", "ibmdos200").unwrap();
        let result = partition.set_start_lba(&disk, 5);

        assert!(result.is_err());
        assert_eq!(result.err().unwrap().kind(), std::io::ErrorKind::Other);
    }

    #[test]
    fn test_set_status_active() {
        let mut partition = Partition::new();
        partition.set_status("active");
        assert_eq!(partition.status, PartitionStatus::Active);
    }

    #[test]
    fn test_set_status_inactive() {
        let mut partition = Partition::new();
        partition.set_status("inactive");
        assert_eq!(partition.status, PartitionStatus::Inactive);
    }

    #[test]
    fn test_set_status_invalid() {
        let mut partition = Partition::new();
        partition.set_status("invalid");
        assert_eq!(partition.status, PartitionStatus::Invalid);
    }

    #[test]
    fn test_set_partition_type_fat12() {
        let mut partition = Partition::new();
        partition.set_partition_type(0x01);
        assert_eq!(partition.partition_type, PartitionType::Fat12MBR);
    }

    #[test]
    fn test_set_partition_type_fat16() {
        let mut partition = Partition::new();
        partition.set_partition_type(0x06); // Assuming 0x06 represents FAT16
        assert_eq!(partition.partition_type, PartitionType::Fat16B);
    }

    #[test]
    fn test_set_partition_type_invalid() {
        let mut partition = Partition::new();
        partition.set_partition_type(0xFF);
        assert_eq!(partition.partition_type, PartitionType::Empty);
    }

    #[test]
    fn test_as_partition_entry_active_fat12() {
        let partition = Partition {
            status: PartitionStatus::Active,
            start_sector: CHS::new(1, 2, 3).unwrap(),
            partition_type: PartitionType::Fat12MBR,
            end_sector: CHS::new(4, 5, 6).unwrap(),
            start_sector_lba: 10,
            sector_count: 20,
            filesystem: None,
        };

        let partition_entry = partition.as_partition_entry();

        assert_eq!(partition_entry[0], PartitionStatus::Active.as_byte());
        assert_eq!(
            &partition_entry[1..4],
            &partition.start_sector.as_mbr_bytes()
        );
        assert_eq!(partition_entry[4], PartitionType::Fat12MBR.as_byte());
        assert_eq!(&partition_entry[5..8], &partition.end_sector.as_mbr_bytes());
        assert_eq!(
            u32::from_le_bytes(partition_entry[12..16].try_into().unwrap()),
            20
        );
    }

    #[test]
    fn test_as_partition_entry_inactive_fat16() {
        let partition = Partition {
            status: PartitionStatus::Inactive,
            start_sector: CHS::new(10, 20, 30).unwrap(),
            partition_type: PartitionType::Fat16MBR,
            end_sector: CHS::new(40, 50, 60).unwrap(),
            start_sector_lba: 100,
            sector_count: 200,
            filesystem: None,
        };

        let partition_entry = partition.as_partition_entry();

        assert_eq!(partition_entry[0], PartitionStatus::Inactive.as_byte());
        assert_eq!(
            &partition_entry[1..4],
            &partition.start_sector.as_mbr_bytes()
        );
        assert_eq!(partition_entry[4], PartitionType::Fat16MBR.as_byte());
        assert_eq!(&partition_entry[5..8], &partition.end_sector.as_mbr_bytes());
        assert_eq!(
            u32::from_le_bytes(partition_entry[12..16].try_into().unwrap()),
            200
        );
    }

    #[test]
    fn test_get_filesystem_with_none() {
        let partition = Partition::new();
        let filesystem = partition.get_filesystem();
        assert_eq!(filesystem, None);
    }

    #[test]
    fn test_as_bytes_active_partition() {
        let partition = Partition {
            status: PartitionStatus::Active,
            start_sector: CHS::new(1, 2, 3).unwrap(),
            partition_type: PartitionType::Fat12MBR,
            end_sector: CHS::new(4, 5, 6).unwrap(),
            start_sector_lba: 10,
            sector_count: 20,
            filesystem: None,
        };

        let partition_entry = partition.as_bytes();

        assert_eq!(partition_entry[0], PartitionStatus::Active.as_byte());
        assert_eq!(
            &partition_entry[1..4],
            &partition.start_sector.as_mbr_bytes()
        );
        assert_eq!(partition_entry[4], PartitionType::Fat12MBR.as_byte());
        assert_eq!(&partition_entry[5..8], &partition.end_sector.as_mbr_bytes());
        assert_eq!(
            u32::from_le_bytes(partition_entry[8..12].try_into().unwrap()),
            10
        );
        assert_eq!(
            u32::from_le_bytes(partition_entry[12..16].try_into().unwrap()),
            20
        );
    }

    #[test]
    fn test_as_bytes_inactive_partition() {
        let partition = Partition {
            status: PartitionStatus::Inactive,
            start_sector: CHS::new(10, 20, 30).unwrap(),
            partition_type: PartitionType::Fat16MBR,
            end_sector: CHS::new(40, 50, 60).unwrap(),
            start_sector_lba: 100,
            sector_count: 200,
            filesystem: None,
        };

        let partition_entry = partition.as_bytes();

        assert_eq!(partition_entry[0], PartitionStatus::Inactive.as_byte());
        assert_eq!(
            &partition_entry[1..4],
            &partition.start_sector.as_mbr_bytes()
        );
        assert_eq!(partition_entry[4], PartitionType::Fat16MBR.as_byte());
        assert_eq!(&partition_entry[5..8], &partition.end_sector.as_mbr_bytes());
        assert_eq!(
            u32::from_le_bytes(partition_entry[8..12].try_into().unwrap()),
            100
        );
        assert_eq!(
            u32::from_le_bytes(partition_entry[12..16].try_into().unwrap()),
            200
        );
    }

    #[test]
    fn test_from_bytes_valid_data() {
        let bytes: [u8; 16] = [
            0x80, // Active status byte
            1, 2, 3,    // Start sector CHS bytes
            0x07, // FAT32 partition type
            4, 5, 6, // End sector CHS bytes
            10, 0, 0, 0, // Start sector LBA address
            20, 0, 0, 0, // Sector count
        ];

        let result = Partition::from_bytes(&bytes).unwrap();
        assert_eq!(result.status, PartitionStatus::Active);
        assert_eq!(result.start_sector, CHS::new(3, 1, 2).unwrap());
        assert_eq!(result.partition_type, PartitionType::Empty);
        assert_eq!(result.end_sector, CHS::new(6, 4, 5).unwrap());
        assert_eq!(result.start_sector_lba, 10);
        assert_eq!(result.sector_count, 20);
    }
}
