use std::fmt;

use super::{partition::Partition, Disk};

#[derive(Debug)]
pub struct PartitionTable {
    pub partitions: [Option<Partition>; 4],
}

impl Default for PartitionTable {
    fn default() -> Self {
        Self::new()
    }
}

//const ARRAY_REPEAT_PARTITION: std::option::Option<Partition> = None;
impl PartitionTable {
    /// Creates a new empty `PartitionTable` with 4 slots.
    ///
    /// The `PartitionTable` is initially empty, with all slots set to `None`.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::PartitionTable;
    ///
    /// let partition_table = PartitionTable::new();
    /// assert_eq!(partition_table.partitions.len(), 4);
    /// ```
    pub fn new() -> Self {
        const ARRAY_REPEAT_PARTITION: std::option::Option<Partition> = None;
        PartitionTable {
            partitions: [ARRAY_REPEAT_PARTITION; 4],
        }
    }

    /// Adds a partition to the partition table.
    ///
    /// If a specific `slot_number` is provided (0 to 3), the partition will be
    /// written into that slot if it's available. If no slot number is provided,
    /// the partition will be written into the lowest available slot number.
    ///
    /// # Arguments
    ///
    /// * `partition` - The partition to add to the partition table.
    /// * `slot_number` - Optional slot number (0 to 3) where the partition should be written.
    ///
    /// # Errors
    ///
    /// Returns an `io::Error` if there's an error adding the partition:
    /// - If the provided slot number is out of range (not in the range 0 to 3).
    /// - If the provided slot is already occupied by another partition.
    /// - If all slots in the partition table are already occupied.
    pub fn add_partition(
        &mut self,
        partition: Partition,
        slot_number: Option<usize>,
    ) -> Result<(), std::io::Error> {
        // If a specific slot number is provided, try to write the partition into that slot
        if let Some(slot) = slot_number {
            if slot >= self.partitions.len() {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "Slot number is out of range",
                ));
            }

            if self.partitions[slot].is_some() {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::AlreadyExists,
                    "Slot is already occupied",
                ));
            }

            self.partitions[slot] = Some(partition);
            return Ok(());
        }

        // If no slot number is provided, find the lowest available slot and write the partition into it
        for slot in self.partitions.iter_mut() {
            if slot.is_none() {
                *slot = Some(partition);
                return Ok(());
            }
        }

        Err(std::io::Error::new(
            std::io::ErrorKind::AlreadyExists,
            "All slots are occupied",
        ))
    }

    /// Converts the partitions into a byte array representation.
    ///
    /// Returns a byte array representing the partitions.
    pub fn as_bytes(&self) -> [u8; 48] {
        let mut partition_table = [0u8; 48];
        let mut offset = 0;
        for partition in &self.partitions {
            if let Some(partition) = partition {
                partition_table[offset..offset + 16].copy_from_slice(&partition.as_bytes())
            };
            offset += 16;
        }
        partition_table
    }

    /// Loads the partition table from the specified disk.
    ///
    /// Reads the partition table from the first sector of the disk and populates
    /// the `partitions` array with the parsed partitions.
    ///
    /// # Arguments
    ///
    /// * `disk` - A mutable reference to the disk from which to load the partition table.
    ///
    /// # Errors
    ///
    /// Returns an `io::Error` if there's an error reading the disk or parsing the partition table.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::{Disk, PartitionTable};
    ///
    /// let mut disk = Disk::new(); // Create a new disk
    /// let mut partition_table = PartitionTable::new(); // Create a new partition table
    ///
    /// partition_table.load(&mut disk).unwrap(); // Load the partition table from the disk
    /// ```
    pub fn load(&mut self, disk: &mut Disk) -> Result<(), std::io::Error> {
        // Read the first sector of the disk
        let buffer = disk.read_sector(0)?;

        // Parse each partition entry from the partition table
        let mut offset = 446;
        for index in 0..4 {
            // Extract the bytes representing the partition entry
            let mut partition_bytes = [0u8; 16];
            partition_bytes.copy_from_slice(&buffer[offset..offset + 16]);

            // Parse the partition entry into a Partition object
            let partition = Partition::from_bytes(&partition_bytes)?;

            // Store the parsed partition in the partitions array
            self.partitions[index] = Some(partition);

            // Move to the next partition entry
            offset += 16;
        }
        Ok(())
    }
}

impl fmt::Display for PartitionTable {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        // Iterate over each partition in the partition table
        for (index, partition) in self.partitions.iter().enumerate() {
            // Check if the partition is Some
            if let Some(partition) = partition {
                // If the partition is Some, print its index and contents
                write!(f, "Partition {}:\n------------\n{}\n\n", index, partition)?;
            } else {
                // If the partition is None, indicate that the slot is empty
                write!(f, "Partition {} is empty\n\n", index)?;
            }
        }

        // Indicate the end of the partition table
        write!(f, "-----------------------------------------------")
    }
}
