use crate::disk::CHS;
use chrono::{DateTime, TimeZone, Utc};
use core::fmt;
use std::fs::File;
use std::io::{Read, Seek, SeekFrom};
use std::str;
//use chrono::format;
use uuid::Uuid;

/// Data structure for the VHD metadata from a disk image's footer
#[derive(PartialEq, Debug)]
pub struct Vhd {
    cookie: String,
    features: VhdFeatures,
    version: u32,
    data_offset: u64,
    timestamp: VhdTimestamp,
    creator_app: String,
    creator_version: u32,
    creator_os: VhdOs,
    original_size: u64,
    current_size: u64,
    geometry: CHS,
    disk_type: VhdDisktype,
    checksum: u32,
    uuid: Uuid,
    saved_state: bool,
}

impl Default for Vhd {
    fn default() -> Self {
        Self::new()
    }
}

impl fmt::Display for Vhd {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.disk_type {
            VhdDisktype::NonVHD => {
                writeln!(f, "Non-VHD or invalid disk type.")
            }
            _ => {
                let offset_display = match self.data_offset {
                    4294967295 => "0".to_string(),
                    _ => self.data_offset.to_string(),
                };
                write!(
                    f,
                    "Features           : {}\n\
             VHD Format version : {}\n\
             Data offset        : {}\n\
             Timestamp          : {}\n\
             Creator app        : {}\n\
             Creator version    : {}\n\
             Creator OS         : {}\n\
             Original size      : {}\n\
             Current size       : {}\n\
             Geometry           : {}\n\
             Disk type          : {}\n\
             Checksum           : {}\n\
             Uuid               : {}\n\
             Saved state        : {}\n",
                    self.features,
                    self.version,
                    offset_display,
                    self.timestamp,
                    self.get_creator(),
                    self.creator_version,
                    self.creator_os,
                    self.original_size,
                    self.current_size,
                    self.geometry,
                    self.disk_type,
                    self.checksum,
                    self.uuid,
                    self.saved_state,
                )
            }
        }
    }
}
impl Vhd {
    /// Constructs a new instance of `Vhd` with default settings.
    ///
    /// This method initializes a new `Vhd` struct with default values for each field. It sets the `cookie` field to the VHD header cookie, `features` to indicate no special features, `version` to the VHD format version, `data_offset` to an invalid value (to be updated later), `timestamp` to the current system time, `creator_app` to the name of the application creating the VHD, `creator_version` to the version number of the creator application, `creator_os` to indicate the operating system of the creator, `original_size` and `current_size` to zero (to be updated later), `geometry` to an empty CHS structure, `disk_type` to indicate a fixed-size hard disk, `checksum` to zero (to be updated later), `uuid` to a randomly generated UUID, and `saved_state` to `false` (indicating not saved state).
    ///
    /// The `checksum` and `saved_state` fields are initialized to zero and `false` respectively, as they may be updated or set by other methods after construction.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let vhd = Vhd::new();
    /// ```
    ///
    /// This example constructs a new `Vhd` instance with default settings.
    pub fn new() -> Self {
        let mut new_vhd = Vhd {
            cookie: Vhd::cookie("conectix".to_string()),
            features: VhdFeatures::NoFeatures,
            version: 0x00010000,
            data_offset: 0xffffffff,
            timestamp: VhdTimestamp::new(),
            creator_app: "dsk8".to_string(),
            creator_version: 1,
            creator_os: VhdOs::Windows,
            original_size: 0,
            current_size: 0,
            geometry: CHS::empty(),
            disk_type: VhdDisktype::FixedHdd,
            checksum: 0, // May be set later
            uuid: Uuid::new_v4(),
            saved_state: false, // May be set later
        };
        new_vhd.set_checksum();
        new_vhd
    }

    /// Returns a clone of the `VhdTimestamp` associated with this VHD.
    ///
    /// This method retrieves a clone of the `VhdTimestamp` associated with the current VHD instance.
    /// The returned `VhdTimestamp` represents the creation time of the VHD and provides information
    /// about the date and time when the VHD was created or modified.
    ///
    /// # Returns
    ///
    /// A clone of the `VhdTimestamp` associated with this VHD.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let vhd = Vhd::new();
    /// let timestamp = vhd.get_timestamp();
    /// println!("VHD creation timestamp: {}", timestamp);
    /// ```
    ///
    /// This example demonstrates how to retrieve the timestamp associated with a VHD instance.
    /// The timestamp is then printed to the console for informational purposes.
    pub fn get_timestamp(&self) -> VhdTimestamp {
        self.timestamp.clone()
    }

    pub fn get_type(&self) -> &VhdDisktype {
        &self.disk_type
    }

    /// Returns the human-readable name of the creator based on the `creator_app` field.
    ///
    /// This method retrieves the human-readable name of the creator based on the value of the `creator_app` field of the VHD.
    /// It maps known creator application identifiers to their corresponding names for better readability and understanding.
    /// If the `creator_app` field does not match any known identifier, the original value of `creator_app` is returned.
    ///
    /// # Returns
    ///
    /// A string representing the human-readable name of the creator.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let vhd = Vhd {
    ///     creator_app: "vbox".to_string(),
    ///     ..Default::default()
    /// };
    ///
    /// assert_eq!(vhd.get_creator(), "Oracle VirtualBox");
    /// ```
    ///
    /// This example demonstrates how to retrieve the human-readable name of the creator based on the `creator_app` field.
    /// In this case, the `creator_app` field is set to "vbox", which corresponds to "Oracle VirtualBox".
    /// The returned string represents the name of the creator.
    pub fn get_creator(&self) -> String {
        match self.creator_app.as_str() {
            "vbox" => "Oracle VirtualBox".to_string(),
            "mvhd" => "PCEm PC Emulator v17 (libvhd)".to_string(),
            "qemu" => "QEMU".to_string(),
            "dsk8" => "DOSContainer".to_string(),
            "win" => "Windows Disk Management".to_string(),
            _ => self.creator_app.clone(), // or self.creator_app.into()
        }
    }

    /// Retrieves the physical geometry of the VHD.
    ///
    /// This method returns a shared reference to the physical geometry (`CHS`) of the VHD.
    /// The geometry represents the cylinder-head-sector layout of the disk.
    ///
    /// # Returns
    ///
    /// A reference to the physical geometry (`CHS`) of the VHD.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let vhd = Vhd::new();
    /// let geometry = vhd.get_geometry();
    /// println!("VHD Geometry: {:?}", geometry);
    /// ```
    ///
    /// This example demonstrates how to retrieve the physical geometry of a VHD instance.
    /// The geometry information is then printed to the console for informational purposes.
    pub fn get_geometry(&self) -> &CHS {
        &self.geometry
    }

    pub fn get_currentsize(&self) -> u64 {
        self.current_size
    }

    /// Convert the `Vhd` struct into a sequence of bytes for writing to disk.
    ///
    /// This method serializes the fields of the `Vhd` struct into a byte sequence
    /// according to the VHD disk format specification. The resulting byte sequence
    /// can be written directly to a disk file.
    ///
    /// # Returns
    ///
    /// A `Vec<u8>` containing the serialized representation of the `Vhd` struct.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let vhd = Vhd::new(); // Create a new VHD instance
    /// let bytes = vhd.as_bytes(); // Convert the VHD to bytes
    /// // Now `bytes` can be written to a disk file
    /// ```
    ///
    /// This example demonstrates how to convert a `Vhd` struct into a byte sequence
    /// for writing to disk. The `as_bytes` method is called on a `Vhd` instance,
    /// producing a byte vector that can be further processed or written to a file.
    pub fn as_bytes(&self) -> Vec<u8> {
        let mut buffer = Vec::with_capacity(512);
    
        // Write each field to the buffer
        buffer.extend_from_slice(self.cookie.as_bytes());
        buffer.extend_from_slice(&self.features.as_bytes());
        buffer.extend_from_slice(&self.version.to_be_bytes());
        buffer.extend_from_slice(&self.data_offset.to_be_bytes());
        buffer.extend_from_slice(&self.timestamp.get_value().to_be_bytes());
        buffer.extend_from_slice(self.creator_app.as_bytes());
        buffer.extend_from_slice(&self.creator_version.to_be_bytes());
        buffer.extend_from_slice(&self.creator_os.as_bytes());
        buffer.extend_from_slice(&self.original_size.to_be_bytes());
        buffer.extend_from_slice(&self.current_size.to_be_bytes());
        buffer.extend_from_slice(&self.geometry.as_vhd_bytes());
        buffer.extend_from_slice(&self.disk_type.as_bytes());
        buffer.extend_from_slice(&self.checksum.to_be_bytes());
        buffer.extend_from_slice(self.uuid.as_bytes());
        buffer.push(if self.saved_state { 1 } else { 0 });
    
        buffer
    }

    /// Ensures the entered string matches the VHD specification and sets the cookie.
    ///
    /// The VHD specification mandates that the cookie field must be 8 characters long. If the
    /// provided string is longer than 8 characters, it is truncated to 8 characters. If it's
    /// shorter, it is padded with spaces until it reaches 8 characters. The final string is then
    /// converted to lowercase.
    ///
    /// # Arguments
    ///
    /// * `cookie` - The string to be validated and set as the VHD cookie.
    ///
    /// # Returns
    ///
    /// A string representing the valid VHD cookie.
    pub fn cookie(mut cookie: String) -> String {
        // Truncate or pad the string to 8 characters
        if cookie.len() > 8 {
            cookie.truncate(8);
        } else {
            cookie.extend(std::iter::repeat(' ').take(8 - cookie.len()));
        }
        cookie.to_ascii_lowercase()
    }

    /// Reconstitute a `Vhd` object from a sequence of bytes.
    ///
    /// # Arguments
    ///
    /// * `buffer` - A slice containing the bytes representing the VHD.
    ///
    /// # Returns
    ///
    /// A `Vhd` object parsed from the provided byte sequence.
    ///
    /// # Errors
    ///
    /// This function returns an `io::Error` if:
    /// * The length of the buffer is less than 78 bytes, which is the minimum required for a valid VHD.
    /// * The provided byte sequence does not represent a valid VHD. This could happen if the first 8 bytes,
    ///   which represent the VHD cookie, do not match the expected value ("conectix").
    ///
    /// # Panics
    ///
    /// This function panics if any conversion from bytes to numbers fails due to invalid byte sequences.
    ///
    /// # Example
    ///
    /// ```
    /// use std::fs::File;
    /// use std::io::{self, Read};
    /// use disk::vhd::Vhd;
    ///
    /// // Read VHD data from a file
    /// fn read_vhd_data(file: &mut File) -> io::Result<Vec<u8>> {
    ///     let mut buffer = Vec::new();
    ///     file.read_to_end(&mut buffer)?;
    ///     Ok(buffer)
    /// }
    ///
    /// // Load a VHD from a file
    /// fn load_vhd(file_path: &str) -> io::Result<Vhd> {
    ///     let mut file = File::open(file_path)?;
    ///     let vhd_data = read_vhd_data(&mut file)?;
    ///     Ok(Vhd::from_bytes(&vhd_data))
    /// }
    ///
    /// // Example usage
    /// let vhd_file_path = "example.vhd";
    /// match load_vhd(vhd_file_path) {
    ///     Ok(vhd) => println!("Loaded VHD: {:?}", vhd),
    ///     Err(e) => eprintln!("Error loading VHD: {}", e),
    /// }
    /// ```
    pub fn from_bytes(buffer: &[u8]) -> Result<Self, std::io::Error> {
        if buffer.len() < 78 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Buffer too small",
            ));
        }

        let to_err = |err: &'static str| -> std::io::Error {
            std::io::Error::new(std::io::ErrorKind::InvalidData, err)
        };

        Ok(Vhd {
            cookie: str::from_utf8(&buffer[0..8])
                .map_err(|_| to_err("Invalid cookie"))?
                .trim_end_matches(char::from(0))
                .to_lowercase(),
            features: VhdFeatures::from_bytes(&buffer[8..12]),
            version: u32::from_be_bytes(
                buffer[12..16]
                    .try_into()
                    .map_err(|_| to_err("Invalid version"))?,
            ),
            data_offset: u64::from_be_bytes(
                buffer[16..24]
                    .try_into()
                    .map_err(|_| to_err("Invalid data offset"))?,
            ),
            timestamp: VhdTimestamp::from_be_bytes(
                buffer[24..28]
                    .try_into()
                    .map_err(|_| to_err("Invalid timestamp"))?,
            ),
            creator_app: str::from_utf8(&buffer[28..32])
                .map_err(|_| to_err("Invalid creator app"))?
                .trim_end_matches(char::from(0))
                .to_string(),
            creator_version: u32::from_be_bytes(
                buffer[32..36]
                    .try_into()
                    .map_err(|_| to_err("Invalid creator version"))?,
            ),
            creator_os: VhdOs::from_bytes(&buffer[36..40]),
            original_size: u64::from_be_bytes(
                buffer[40..48]
                    .try_into()
                    .map_err(|_| to_err("Invalid original size"))?,
            ),
            current_size: u64::from_be_bytes(
                buffer[48..56]
                    .try_into()
                    .map_err(|_| to_err("Invalid current size"))?,
            ),
            geometry: CHS::from_vhd_bytes(
                buffer[56..60]
                    .try_into()
                    .map_err(|_| to_err("Invalid geometry"))?,
            ),
            disk_type: match str::from_utf8(&buffer[0..8]).unwrap() {
                "conectix" => VhdDisktype::from_bytes(&buffer[60..64]),
                _ => VhdDisktype::NonVHD,
            },
            checksum: u32::from_be_bytes(
                buffer[64..68]
                    .try_into()
                    .map_err(|_| to_err("Invalid checksum"))?,
            ),
            uuid: Uuid::from_slice(&buffer[68..84]).map_err(|_| to_err("Invalid UUID"))?,
            saved_state: buffer[83] == 1,
        })
    }

    /// Load a `Vhd` object from the end of a file.
    ///
    /// Reads the last 512 bytes of the file into memory and parses them into a `Vhd` object.
    ///
    /// # Arguments
    ///
    /// * `file` - A mutable reference to the file to load the VHD from.
    ///
    /// # Returns
    ///
    /// A `Result` containing the loaded `Vhd` object if successful, or an `std::io::Error` if an I/O error occurs.
    ///
    /// # Errors
    ///
    /// Returns an `std::io::Error` if:
    /// * Seeking to the starting position fails.
    /// * Reading the last 512 bytes into the buffer fails.
    /// * Parsing the bytes into a `Vhd` object fails.
    ///
    /// # Example
    ///
    /// ```no_run
    /// use std::fs::File;
    /// use std::io;
    /// use disk::vhd::Vhd;
    ///
    /// // Function to load a VHD from a file
    /// fn load_vhd_from_file(file_path: &str) -> io::Result<Vhd> {
    ///     let mut file = File::open(file_path)?;
    ///     Vhd::load(&mut file)
    /// }
    ///
    /// // Example usage
    /// match load_vhd_from_file("example.vhd") {
    ///     Ok(vhd) => println!("Loaded VHD: {:?}", vhd),
    ///     Err(err) => eprintln!("Error loading VHD: {}", err),
    /// }
    /// ```
    pub fn load(file: &mut File) -> Result<Self, std::io::Error> {
        // Get the length of the file
        let file_size = file.seek(SeekFrom::End(0))?;

        // If the file is empty, return an empty VHD object
        if file_size == 0 {
            return Ok(Vhd::new()); // Assuming Vhd::new() creates an empty VHD object
        }

        // Calculate the starting position to read from
        let start_position = if file_size > 512 { file_size - 512 } else { 0 };

        // Seek to the starting position
        file.seek(SeekFrom::Start(start_position))?;

        // Read the last 512 bytes into a buffer
        let mut buffer = [0; 512];
        let bytes_read = file.read(&mut buffer)?;

        // If the bytes read is less than 512, pad the buffer with zeros
        if bytes_read < buffer.len() {
            for i in bytes_read..buffer.len() {
                buffer[i] = 0;
            }
        }

        Vhd::from_bytes(&buffer)
    }

    /// Sets the geometry of the VHD disk.
    ///
    /// # Arguments
    ///
    /// * `geometry` - A reference to a `CHS` struct representing the geometry
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::Vhd;
    /// use your_crate_name::CHS;
    ///
    /// let mut vhd = Vhd::new();
    /// let geometry = CHS {
    ///     cylinder: 100,
    ///     head: 10,
    ///     sector: 5,
    /// };
    /// vhd.set_geometry(geometry);
    /// ```
    pub fn set_geometry(&mut self, geometry: CHS) {
        self.geometry = geometry;
        self.set_checksum();
    }

    /// Sets the size of the VHD disk.
    ///
    /// # Arguments
    ///
    /// * `size` - The new size of the VHD disk in bytes. Limited to a maximum of 4GB.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let mut vhd = Vhd::new();
    /// vhd.set_size(1024);
    /// ```
    pub fn set_size(&mut self, size: u32) {
        // Convert size to u64 to adhere to the 64-bit file format specification
        let size_u64 = size as u64;
        // Update current_size and original_size
        self.current_size = size_u64;
        self.original_size = size_u64;
        // Recalculate checksum
        self.set_checksum();
    }

    /// Calculates and sets the checksum value for the VHD metadata.
    ///
    /// This method computes the checksum value based on the VHD metadata fields
    /// and sets it in the `checksum` field of the VHD struct. The checksum is
    /// calculated as the bitwise complement of the sum of all bytes in the metadata.
    ///
    /// # Example
    ///
    /// ```
    /// use your_crate_name::Vhd;
    ///
    /// let mut vhd = Vhd::new();
    /// vhd.set_checksum();
    /// ```
    pub fn set_checksum(&mut self) {
        let mut buffer: Vec<u8> = Vec::with_capacity(512);
        buffer.extend_from_slice(self.cookie.as_bytes());
        buffer.extend_from_slice(&self.features.as_bytes());
        buffer.extend_from_slice(&self.version.to_be_bytes());
        buffer.extend_from_slice(&self.data_offset.to_be_bytes());
        buffer.extend_from_slice(&self.timestamp.to_be_bytes());
        buffer.extend_from_slice(self.creator_app.as_bytes());
        buffer.extend_from_slice(&self.creator_version.to_be_bytes());
        buffer.extend_from_slice(&self.creator_os.as_bytes());
        buffer.extend_from_slice(&self.original_size.to_be_bytes());
        buffer.extend_from_slice(&self.current_size.to_be_bytes());
        buffer.extend_from_slice(&self.geometry.as_vhd_bytes());
        buffer.extend_from_slice(&self.disk_type.as_bytes());
        buffer.extend_from_slice(&[0u8; 4]); // This is where checksum would regularly be
        buffer.extend_from_slice(self.uuid.as_bytes());
        if self.saved_state {
            buffer.push(1);
        } else {
            buffer.push(0);
        }
        let sum: u32 = buffer.iter().map(|&x| x as u32).sum();
        self.checksum = !sum;
    }
}

#[derive(Clone, PartialEq)]
pub struct VhdTimestamp {
    timestamp: u32,
}

impl Default for VhdTimestamp {
    fn default() -> Self {
        Self::new()
    }
}

impl VhdTimestamp {
    pub fn new() -> Self {
        let start_of_2000 = Utc.with_ymd_and_hms(2000, 1, 1, 0, 0, 0).unwrap();
        let current_time = Utc::now();
        let duration_since_2000 = current_time.signed_duration_since(start_of_2000);
        Self {
            timestamp: duration_since_2000.num_seconds() as u32,
        }
    }

    pub fn get_value(&self) -> u32 {
        self.timestamp
    }

    pub fn to_be_bytes(&self) -> [u8; 4] {
        u32::to_be_bytes(self.timestamp)
    }

    pub fn from_be_bytes(bytes: &[u8; 4]) -> Self {
        Self {
            timestamp: u32::from_be_bytes(*bytes),
        }
    }
}

impl fmt::Display for VhdTimestamp {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let start_of_2000: DateTime<Utc> = Utc.with_ymd_and_hms(2000, 1, 1, 0, 0, 0).unwrap();
        let timestamp_duration = chrono::TimeDelta::try_seconds(self.timestamp as i64).unwrap();
        let date_time = start_of_2000 + timestamp_duration;

        let target_timestamp = date_time.to_rfc3339();
        write!(f, "{}", target_timestamp)
    }
}

impl fmt::Debug for VhdTimestamp {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt::Display::fmt(self, f)
    }
}

#[derive(PartialEq, Debug, Default)]
pub enum VhdFeatures {
    #[default]
    NoFeatures,
    Temporary,
    Reserved,
}

impl fmt::Display for VhdFeatures {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let display_features = match self {
            Self::NoFeatures => "No features",
            Self::Reserved => "Reserved",
            Self::Temporary => "Temporary",
        };
        write!(f, "{}", display_features)
    }
}

impl VhdFeatures {
    pub fn as_bytes(&self) -> [u8; 4] {
        match self {
            Self::NoFeatures => [0, 0, 0, 0],
            Self::Temporary => [0, 0, 0, 1],
            Self::Reserved => [0, 0, 0, 2],
        }
    }

    pub fn from_bytes(bytes: &[u8]) -> Self {
        match bytes {
            [0, 0, 0, 0] => Self::NoFeatures,
            [0, 0, 0, 1] => Self::Temporary,
            [0, 0, 0, 2] => Self::Reserved,
            _ => Self::NoFeatures,
        }
    }
}

#[derive(PartialEq, Debug)]
pub enum VhdOs {
    Windows,
    Macintosh,
}

impl fmt::Display for VhdOs {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let display_os = match self {
            Self::Macintosh => "Macintosh",
            Self::Windows => "Windows",
        };
        write!(f, "{}", display_os)
    }
}

impl VhdOs {
    /// Return the OS as an u32 value
    pub fn as_u32(&self) -> u32 {
        match self {
            Self::Windows => 0x5769326b,
            Self::Macintosh => 0x4d616320,
        }
    }

    /// Return the OS as on-disk big-endian bytes
    pub fn as_bytes(&self) -> [u8; 4] {
        self.as_u32().to_be_bytes()
    }

    /// Compose object from on-disk bytes, Mac for Mac, Windows for all others.
    pub fn from_bytes(bytes: &[u8]) -> Self {
        let number = u32::from_be_bytes(
            bytes
                .try_into()
                .expect("Failed to convert bytes to number."),
        );
        match number {
            0x4d616320 => Self::Macintosh,
            _ => Self::Windows,
        }
    }
}

#[derive(Debug, PartialEq)]
pub enum VhdDisktype {
    None,
    Reserved,
    FixedHdd,
    DynamicHdd,
    DifferencingHdd,
    NonVHD,
}

impl fmt::Display for VhdDisktype {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let display_disktype = match self {
            Self::None => "None",
            Self::Reserved => "Reserved value",
            Self::FixedHdd => "Fixed hard disk",
            Self::DynamicHdd => "Dynamic hard disk",
            Self::DifferencingHdd => "Differencing hard disk",
            Self::NonVHD => "Not a VHD format disk",
        };
        write!(f, "{}", display_disktype)
    }
}

/// Types 5 and 6 are just as deprecated as 1 so skip them and use 1.
impl VhdDisktype {
    pub fn as_u32(&self) -> u32 {
        match self {
            Self::None => 0,
            Self::Reserved => 1,
            Self::FixedHdd => 2,
            Self::DynamicHdd => 3,
            Self::DifferencingHdd => 4,
            Self::NonVHD => 7,
        }
    }

    /// Return the VHD Disk type as on-disk big-endian bytes
    pub fn as_bytes(&self) -> [u8; 4] {
        self.as_u32().to_be_bytes()
    }

    /// Reconstruct the VHD Disk type from on-disk bytes
    pub fn from_bytes(bytes: &[u8]) -> Self {
        let number = u32::from_be_bytes(
            bytes
                .try_into()
                .expect("Failed to convert bytes to number."),
        );
        match number {
            0 => Self::None,
            2 => Self::FixedHdd,
            3 => Self::DynamicHdd,
            4 => Self::DifferencingHdd,
            7 => Self::NonVHD,
            _ => Self::Reserved,
        }
    }
}

#[cfg(test)]
mod tests {

    use super::*;

    #[test]
    fn test_new_default_values() {
        // Create a new VHD instance
        let vhd = Vhd::new();

        // Check default values
        assert_eq!(vhd.cookie, Vhd::cookie("conectix".to_string()));
        assert_eq!(vhd.features, VhdFeatures::NoFeatures);
        assert_eq!(vhd.version, 0x00010000);
        assert_eq!(vhd.data_offset, 0xffffffff);
        // Timestamp comparison may vary, so we don't assert it
        assert_eq!(vhd.creator_app, "dsk8".to_string());
        assert_eq!(vhd.creator_version, 1);
        assert_eq!(vhd.creator_os, VhdOs::Windows);
        assert_eq!(vhd.original_size, 0);
        assert_eq!(vhd.current_size, 0);
        assert_eq!(vhd.geometry, CHS::empty());
        assert_eq!(vhd.disk_type, VhdDisktype::FixedHdd);
        // Check UUID is not nil
        assert_ne!(vhd.uuid, Uuid::nil());
        // Check default values for checksum and saved_state
        assert_eq!(vhd.saved_state, false);
    }

    #[test]
    fn test_new_checksum_saved_state() {
        // Create a new VHD instance
        let vhd = Vhd::new();

        // Check default values for checksum and saved_state
        assert_eq!(vhd.saved_state, false);
    }

    #[test]
    fn test_get_timestamp() {
        // Create a new VHD instance
        let vhd = Vhd::new();

        // Retrieve the timestamp
        let timestamp = vhd.get_timestamp();

        // Ensure that the retrieved timestamp matches the timestamp associated with the VHD
        assert_eq!(timestamp, vhd.timestamp);
    }

    #[test]
    fn test_get_creator_vbox() {
        // Create a VHD instance with creator_app set to "vbox"
        let vhd = Vhd {
            creator_app: "vbox".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as "Oracle VirtualBox"
        assert_eq!(vhd.get_creator(), "Oracle VirtualBox");
    }

    #[test]
    fn test_get_creator_mvhd() {
        // Create a VHD instance with creator_app set to "mvhd"
        let vhd = Vhd {
            creator_app: "mvhd".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as "PCEm PC Emulator v17 (libvhd)"
        assert_eq!(vhd.get_creator(), "PCEm PC Emulator v17 (libvhd)");
    }

    #[test]
    fn test_get_creator_qemu() {
        // Create a VHD instance with creator_app set to "qemu"
        let vhd = Vhd {
            creator_app: "qemu".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as "QEMU"
        assert_eq!(vhd.get_creator(), "QEMU");
    }

    #[test]
    fn test_get_creator_windows() {
        // Create a VHD instance with creator_app set to "win"
        let vhd = Vhd {
            creator_app: "win".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as "Windows Disk Management"
        assert_eq!(vhd.get_creator(), "Windows Disk Management");
    }

    #[test]
    fn test_get_creator_dsk8() {
        // Create a VHD instance with creator_app set to "dsk8"
        let vhd = Vhd {
            creator_app: "dsk8".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as "DOSContainer"
        assert_eq!(vhd.get_creator(), "DOSContainer");
    }

    #[test]
    fn test_get_creator_other() {
        // Create a VHD instance with creator_app set to an unknown value
        let vhd = Vhd {
            creator_app: "unknown".to_string(),
            ..Default::default()
        };

        // Check if the creator is correctly identified as the original creator_app value
        assert_eq!(vhd.get_creator(), "unknown");
    }

    #[test]
    fn test_get_geometry() {
        // Create a new VHD instance with a specific geometry
        let geometry = CHS {
            cylinder: 100,
            head: 16,
            sector: 63,
        };
        let vhd = Vhd {
            geometry: geometry.clone(),
            ..Default::default()
        };

        // Retrieve the geometry using get_geometry method
        let retrieved_geometry = vhd.get_geometry();

        // Ensure that the retrieved geometry matches the original geometry
        assert_eq!(retrieved_geometry, &geometry);
    }

    #[test]
    fn test_as_bytes() {
        // Create a sample Vhd instance using the new() method
        let vhd = Vhd::new();

        // Convert the Vhd instance to bytes
        let mut bytes = vhd.as_bytes();

        // Pad the buffer up to capacity
        let current_len = bytes.len();
        let padding_len = 512 - current_len;
        bytes.extend(vec![0; padding_len]);

        // Expected byte sequence generated manually
        let mut expected_bytes: [u8; 512] = [0; 512];
        expected_bytes[0..8].copy_from_slice(vhd.cookie.as_bytes());
        expected_bytes[8..12].copy_from_slice(&vhd.features.as_bytes());
        expected_bytes[12..16].copy_from_slice(&vhd.version.to_be_bytes());
        expected_bytes[16..24].copy_from_slice(&vhd.data_offset.to_be_bytes());
        expected_bytes[24..28].copy_from_slice(&vhd.timestamp.get_value().to_be_bytes());
        expected_bytes[28..32].copy_from_slice(vhd.creator_app.as_bytes());
        expected_bytes[32..36].copy_from_slice(&vhd.creator_version.to_be_bytes());
        expected_bytes[36..40].copy_from_slice(&vhd.creator_os.as_bytes());
        expected_bytes[40..48].copy_from_slice(&vhd.original_size.to_be_bytes());
        expected_bytes[48..56].copy_from_slice(&vhd.current_size.to_be_bytes());
        expected_bytes[56..60].copy_from_slice(&vhd.geometry.as_vhd_bytes());
        expected_bytes[60..64].copy_from_slice(&vhd.disk_type.as_bytes());
        expected_bytes[64..68].copy_from_slice(&vhd.checksum.to_be_bytes());
        expected_bytes[68..84].copy_from_slice(vhd.uuid.as_bytes());
        expected_bytes[511] = if vhd.saved_state { 1 } else { 0 };

        // Compare the generated bytes with the expected byte sequence
        assert_eq!(&bytes[..], &expected_bytes[..]);
    }

    #[test]
    fn test_cookie_truncate_long_string() {
        let long_string = String::from("1234567890");
        assert_eq!(Vhd::cookie(long_string), "12345678");
    }

    #[test]
    fn test_cookie_pad_short_string() {
        let short_string = String::from("abc");
        assert_eq!(Vhd::cookie(short_string), "abc     ");
    }

    #[test]
    fn test_cookie_no_change_correct_length() {
        let correct_length = String::from("abcdefgh");
        assert_eq!(Vhd::cookie(correct_length), "abcdefgh");
    }

    #[test]
    fn test_cookie_no_change_max_length() {
        let max_length = String::from("abcdefghi");
        assert_eq!(Vhd::cookie(max_length), "abcdefgh");
    }

    #[test]
    fn test_cookie_lowercase_conversion() {
        let uppercase = String::from("ABC");
        assert_eq!(Vhd::cookie(uppercase), "abc     ");
    }

    #[test]
    fn test_cookie_empty_string() {
        let empty = String::new();
        assert_eq!(Vhd::cookie(empty), "        ");
    }

    #[test]
    fn test_from_bytes_valid() {
        // Create a valid VHD as_bytes
        let vhd = Vhd::new();
        let vhd_bytes = vhd.as_bytes();

        // Convert it back to Vhd
        let result = Vhd::from_bytes(&vhd_bytes);
        assert!(result.is_ok());
        let parsed_vhd = result.unwrap();

        // Compare original and parsed VHDs
        assert_eq!(vhd, parsed_vhd);
    }

    #[test]
    fn test_from_bytes_invalid_data() {
        // Create invalid VHD data (less than 78 bytes)
        let vhd_bytes = vec![0; 50];

        // Attempt to parse it
        let result = Vhd::from_bytes(&vhd_bytes);
        assert!(result.is_err());
        assert_eq!(
            result.err().unwrap().kind(),
            std::io::ErrorKind::InvalidData
        );
    }

    // Helper function to create a Vhd instance with dummy data
    fn create_dummy_vhd() -> Vhd {
        // Create a Vhd instance with some default values
        Vhd {
            cookie: "conectix".to_string(),
            features: VhdFeatures::default(),
            version: 0x00010000,
            data_offset: 0xffffffff,
            timestamp: VhdTimestamp::default(),
            creator_app: "dsk8".to_string(),
            creator_version: 1,
            creator_os: VhdOs::Windows,
            original_size: 0,
            current_size: 0,
            geometry: CHS::empty(),
            disk_type: VhdDisktype::FixedHdd,
            checksum: 0,
            uuid: Uuid::new_v4(),
            saved_state: false,
        }
    }

    // Test loading a valid VHD file
    #[test]
    fn test_load_valid_vhd() {
        // Create a VHD instance with dummy data
        let dummy_vhd = create_dummy_vhd();

        // Create a temporary file to write the VHD data
        let mut file = tempfile::tempfile().unwrap();
        std::io::Write::write_all(&mut file, &dummy_vhd.as_bytes()).unwrap();

        // Seek to the beginning of the file
        file.seek(SeekFrom::Start(0)).unwrap();

        // Call the load method to load the VHD from the file
        let result = Vhd::load(&mut file);

        // Check that the result is Ok and contains the expected VHD
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), dummy_vhd);
    }

    // Test loading an empty file
    #[test]
    fn test_load_empty_file() {
        // Create a temporary empty file
        let mut file = tempfile::tempfile().unwrap();

        // Call the load method
        let result = Vhd::load(&mut file);

        // Check that the result is an error
        assert!(result.is_ok());
    }

    #[test]
    fn test_set_geometry() {
        // Arrange
        let mut vhd = Vhd::new();
        let geometry = CHS {
            cylinder: 100,
            head: 10,
            sector: 5,
        };

        // Act
        vhd.set_geometry(geometry);

        // Assert
        assert_eq!(vhd.geometry.cylinder, 100);
        assert_eq!(vhd.geometry.head, 10);
        assert_eq!(vhd.geometry.sector, 5);
    }

    #[test]
    fn test_set_geometry_with_zero_values() {
        // Arrange
        let mut vhd = Vhd::new();
        let geometry = CHS {
            cylinder: 0,
            head: 0,
            sector: 0,
        };

        // Act
        vhd.set_geometry(geometry);

        // Assert
        assert_eq!(vhd.geometry.cylinder, 0);
        assert_eq!(vhd.geometry.head, 0);
        assert_eq!(vhd.geometry.sector, 0);
    }

    #[test]
    fn test_set_geometry_with_large_values() {
        // Arrange
        let mut vhd = Vhd::new();
        let geometry = CHS {
            cylinder: 65535,
            head: 255,
            sector: 63,
        };

        // Act
        vhd.set_geometry(geometry);

        // Assert
        assert_eq!(vhd.geometry.cylinder, 65535);
        assert_eq!(vhd.geometry.head, 255);
        assert_eq!(vhd.geometry.sector, 63);
    }

    #[test]
    fn test_set_size() {
        // Arrange
        let mut vhd = Vhd::new();

        // Act
        vhd.set_size(1024);

        // Assert
        assert_eq!(vhd.current_size, 1024);
        assert_eq!(vhd.original_size, 1024);
    }

    #[test]
    fn test_set_size_large_value() {
        // Arrange
        let mut vhd = Vhd::new();

        // Act
        vhd.set_size(u32::MAX);

        // Assert
        assert_eq!(vhd.current_size, u32::MAX as u64);
        assert_eq!(vhd.original_size, u32::MAX as u64);
    }

    #[test]
    fn test_set_size_zero_value() {
        // Arrange
        let mut vhd = Vhd::new();

        // Act
        vhd.set_size(0);

        // Assert
        assert_eq!(vhd.current_size, 0);
        assert_eq!(vhd.original_size, 0);
    }

    #[test]
    fn test_set_checksum() {
        let mut vhd = Vhd::new();
        vhd.set_checksum();

        // Expected checksum value is the bitwise complement of the sum of all bytes in the metadata
        let expected_checksum: u32 = {
            let mut buffer: Vec<u8> = Vec::with_capacity(512);
            buffer.extend_from_slice(vhd.cookie.as_bytes());
            buffer.extend_from_slice(&vhd.features.as_bytes());
            buffer.extend_from_slice(&vhd.version.to_be_bytes());
            buffer.extend_from_slice(&vhd.data_offset.to_be_bytes());
            buffer.extend_from_slice(&vhd.timestamp.to_be_bytes());
            buffer.extend_from_slice(vhd.creator_app.as_bytes());
            buffer.extend_from_slice(&vhd.creator_version.to_be_bytes());
            buffer.extend_from_slice(&vhd.creator_os.as_bytes());
            buffer.extend_from_slice(&vhd.original_size.to_be_bytes());
            buffer.extend_from_slice(&vhd.current_size.to_be_bytes());
            buffer.extend_from_slice(&vhd.geometry.as_vhd_bytes());
            buffer.extend_from_slice(&vhd.disk_type.as_bytes());
            buffer.extend_from_slice(&[0u8; 4]); // Placeholder for checksum
            buffer.extend_from_slice(vhd.uuid.as_bytes());
            if vhd.saved_state {
                buffer.push(1);
            } else {
                buffer.push(0);
            }

            let sum: u32 = buffer.iter().map(|&x| x as u32).sum();
            !sum
        };

        assert_eq!(vhd.checksum, expected_checksum);
    }
}
