use std::{
    fs::{self, File},
    io::{Read, Write},
    path::Path,
};

use indicatif::{ProgressBar, ProgressStyle};
use tempfile::{tempdir, TempDir};
use zip::ZipArchive;

use self::file::DownloadedFile;

pub mod file;

#[derive(Debug)]
pub struct ZipDownloader {
    tempdir: TempDir,
    files: Vec<DownloadedFile>,
}

impl Default for ZipDownloader {
    fn default() -> Self {
        Self::new()
    }
}

impl ZipDownloader {
    pub fn new() -> Self {
        ZipDownloader {
            tempdir: tempdir().unwrap(),
            files: Vec::new(),
        }
    }

    /// Construct a new ZipDownloader from a URL. Combines the construction and initial download
    /// of the ZIP file in a single function call. The struct can then be used to operate on its
    /// contents.
    pub fn from_url(url: &str, message: Option<&str>) -> Result<Self, std::io::Error> {
        let mut downloader = Self::new();
        downloader.download_zip(url, message)?;
        Ok(downloader)
    }

    pub fn get_files(&self) -> &Vec<DownloadedFile> {
        &self.files
    }

    pub fn get_file_by_name(&self, name: &str) -> Option<&DownloadedFile> {
        // Iterate over each DownloadedFile in the vector
        for file in &self.files {
            // Check if the filename matches the provided name
            if file.get_name() == name {
                // If found, return a reference to the DownloadedFile
                return Some(file);
            }
        }
        // If the file is not found, return None
        None
    }

    pub fn download_zip(&mut self, url: &str, message: Option<&str>) -> std::io::Result<()> {
        let temp_dir = &self.tempdir;
        let temp_dir_path = temp_dir.path();

        let zip_path = temp_dir_path.join("downloaded_file.zip");

        if message.is_some() {
            println!("{}", message.unwrap());
        }

        // Download the ZIP file
        Self::download_file(url, &zip_path)?;

        // Extract the archive
        self.extract_zip(&zip_path, temp_dir_path)?;

        // Fill the files vectory
        self.files = Self::gather_files_from_directory(temp_dir_path)?;

        Ok(())
    }

    /// Pull a file from the URL and save it to the file path.
    fn download_file(url: &str, file_path: &Path) -> std::io::Result<()> {
        let mut response = reqwest::blocking::get(url).unwrap();
        let total_size = response.content_length().unwrap_or(0);

        let mut download_file = File::create(file_path)?;

        let pb = ProgressBar::new(total_size);
        pb.set_style(ProgressStyle::default_bar()
            .template("{spinner:.green} [{elapsed_precise}] [{bar:40.cyan/blue}] {bytes}/{total_bytes} ({eta})").unwrap()
            .progress_chars("#>-"));
        pb.set_position(0);

        let mut buf = [0; 8192];
        loop {
            let bytes_read = response.read(&mut buf)?;
            if bytes_read == 0 {
                break;
            }
            download_file.write_all(&buf[..bytes_read])?;
            pb.inc(bytes_read as u64);
        }
        pb.finish_with_message("Download completed.");

        Ok(())
    }

    fn extract_zip(&self, zip_path: &Path, extract_to: &Path) -> std::io::Result<()> {
        let file = File::open(zip_path)?;
        let mut archive = ZipArchive::new(file)?;

        for i in 0..archive.len() {
            let mut file = archive.by_index(i)?;
            let mut output_file = File::create(extract_to.join(file.mangled_name()))?;
            std::io::copy(&mut file, &mut output_file)?;
        }
        std::fs::remove_file(zip_path)?;
        Ok(())
    }

    fn gather_files_from_directory(
        directory_path: &Path,
    ) -> Result<Vec<DownloadedFile>, std::io::Error> {
        let mut downloaded_files = Vec::new();

        // Read the directory
        for entry in fs::read_dir(directory_path)? {
            let entry = entry?;

            // Check if the entry is a file
            if entry.file_type()?.is_file() {
                // Read the file contents
                let file_bytes = fs::read(entry.path())?;

                // Create a DownloadedFile instance with directory set to None
                let downloaded_file =
                    DownloadedFile::new(&entry.file_name().to_string_lossy(), &file_bytes);

                // Add the DownloadedFile instance to the vector
                downloaded_files.push(downloaded_file);
            }
        }

        Ok(downloaded_files)
    }
}
