use std::fmt;

use self::allocationtable::AllocationTable;
use self::direntry::{Arena, DirEntry, DirEntryType, NodeId};
use crate::disk::partition::{Partition, PartitionType};
use crate::disk::vhd::VhdDisktype;
use crate::fat::direntry::Node;
use crate::os::OperatingSystem;
use crate::Disk;

pub mod allocationtable;
pub mod attributes;
pub mod cluster;
pub mod direntry;
pub mod filename;
pub mod timedate;
pub mod vbr;

#[derive(Clone, Debug, PartialEq)]
/// Represents the overall structure of a FAT filesystem.
pub struct Fat {
    /// Represents the type of the FAT filesystem (FAT12, FAT16, FAT32).
    pub fat_type: FatType,
    /// Holds the allocation table, which manages cluster allocation and deallocation.
    pub allocation_table: AllocationTable,
    /// Holds the root directory entries.
    pub arena: Arena<DirEntry>,
    pub root_directory: NodeId,
    /// Stores the OEM name of the filesystem.
    oem_name: String,
    /// Indicates the number of bytes per sector in the filesystem.
    bytes_per_sector: u16,
    /// Specifies the number of sectors per cluster.
    sectors_per_cluster: u8,
    /// Indicates the count of reserved sectors at the beginning of the filesystem.
    pub reserved_sector_count: u16,
    /// Specifies the number of copies of the FAT table.
    pub fat_copies: u8,
    /// Indicates the number of entries in the root directory (only applicable for FAT12 and FAT16).
    root_dir_entries: u16,
    /// Represents the total count of sectors in the filesystem.
    sector_count: u16,
    /// Stores the media descriptor byte, indicating the type of storage media.
    media_descriptor: MediaDescriptor,
    /// Specifies the count of sectors per FAT table.
    pub sectors_per_fat: u16,
    /// Indicates the count of sectors per track.
    sectors_per_track: u16,
    /// Specifies the count of heads (surfaces) on the storage device.
    heads: u16,
    /// Indicates the count of hidden sectors before the partition.
    hidden_sectors: u16,
    /// Holds the bootstrap code or boot sector of the filesystem.
    bootstrap: [u8; 446],
    /// Stores the signature bytes of the filesystem.
    signature: [u8; 2],
    /// Represents the operating system associated with the filesystem.
    os: Option<OperatingSystem>,
}

impl fmt::Display for Fat {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        // Unwrap the Option<OperatingSystem> field
        match &self.os {
            Some(os) => {
                write!(
                    f,
                    "FAT type               : {}\n\
                     OEM Name               : {}\n\
                     Sectors per cluster    : {}\n\
                     Reserved sectors       : {}\n\
                     FAT copies             : {}\n\
                     Root directory entries : {}\n\
                     Sector count           : {}\n\
                     Media descriptor       : {}\n\
                     Sectors per FAT        : {}\n\
                     Hidden sectors         : {}\n\
                     Operating system       : {}\n",
                    self.fat_type,
                    self.oem_name,
                    self.sectors_per_cluster,
                    self.reserved_sector_count,
                    self.fat_copies,
                    self.root_dir_entries,
                    self.sector_count,
                    self.media_descriptor,
                    self.sectors_per_fat,
                    self.hidden_sectors,
                    os // Using the unwrapped value of OperatingSystem
                )
            }
            None => {
                write!(
                    f,
                    "FAT type               : {}\n\
                     OEM Name               : {}\n\
                     Sectors per cluster    : {}\n\
                     Reserved sectors       : {}\n\
                     FAT copies             : {}\n\
                     Root directory entries : {}\n\
                     Sector count           : {}\n\
                     Media descriptor       : {}\n\
                     Sectors per FAT        : {}\n\
                     Hidden sectors         : {}\n\
                     Operating system       : None\n",
                    self.fat_type,
                    self.oem_name,
                    self.sectors_per_cluster,
                    self.reserved_sector_count,
                    self.fat_copies,
                    self.root_dir_entries,
                    self.sector_count,
                    self.media_descriptor,
                    self.sectors_per_fat,
                    self.hidden_sectors,
                )
            }
        }
    }
}

impl Fat {
    /// Constructs a new `Fat` instance based on disk and partition information.
    ///
    /// This method initializes a `Fat` instance by extracting relevant information from the provided `Disk` and `Partition`.
    /// It determines the FAT type based on the partition type, calculates various parameters such as sector count and sectors per cluster,
    /// and initializes the allocation table and other fields accordingly.
    ///
    /// # Arguments
    ///
    /// * `disk` - A reference to a `Disk` representing the storage device containing the FAT filesystem.
    /// * `partition` - A reference to a `Partition` representing the partition containing the FAT filesystem.
    /// * `os` - An optional `OperatingSystem` enum representing the operating system associated with the filesystem.
    ///
    /// # Returns
    ///
    /// A Result containing a new `Fat` instance if initialization succeeds, or an `std::io::Error` if an error occurs.
    ///
    /// # Errors
    ///
    /// This method may return an error if:
    /// * The partition type is unsupported for FAT filesystems.
    /// * Conversion of sector count fails.
    /// * Calculation of sectors per FAT fails.
    ///
    pub fn new(
        disk: &Disk,
        partition: &Partition,
        os: Option<OperatingSystem>,
    ) -> Result<Self, std::io::Error> {
        const SECTOR_SIZE: u16 = 512;
        let geometry = disk.get_geometry()?;
        let fat_type = match partition.get_partition_type() {
            PartitionType::Fat12MBR => FatType::Fat12,
            PartitionType::Fat16MBR => FatType::Fat16,
            PartitionType::Fat16B => FatType::Fat16B,
            _ => {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidData,
                    "Unsupported partition type for FAT filesystem.",
                ));
            }
        };

        let sector_count: u32 = partition.get_sector_count();
        let sectors_per_cluster = Self::calculate_sectors_per_cluster(sector_count, os)?;
        let mbr_bootcode = match os {
            Some(ref os) => os.get_mbr_bootcode(),
            None => [0u8; 446],
        };

        let mut arena = Arena::new();
        let root_directory = arena.insert_node(Node::new_directory("root", Vec::new().as_slice()));
        Ok(Fat {
            fat_type,
            allocation_table: AllocationTable::new(
                sector_count / sectors_per_cluster as u32,
            ),
            arena,
            root_directory,
            oem_name: String::from("IBM  2.0"),
            bytes_per_sector: SECTOR_SIZE,
            sectors_per_cluster,
            reserved_sector_count: 1,
            fat_copies: 2,
            root_dir_entries: 512, // Assuming default value for now
            sector_count: sector_count as u16,
            media_descriptor: MediaDescriptor::HardDrive,
            sectors_per_fat: Self::calculate_sectors_per_fat(
                sector_count,
                sectors_per_cluster as u32,
                fat_type,
            ),
            sectors_per_track: geometry.sector as u16,
            heads: geometry.head as u16,
            hidden_sectors: 0,
            bootstrap: mbr_bootcode,
            signature: [0x55, 0xaa],
            os,
        })
    }

    /// Instantiate the Fat struct from the underlying partition.
    ///
    /// Loads the FAT filesystem metadata from the Volume Boot Record (VBR) located in the given partition on the provided disk.
    ///
    /// # Arguments
    ///
    /// * `disk` - A mutable reference to a `Disk` representing the storage device containing the FAT filesystem.
    /// * `partition` - A reference to a `Partition` representing the partition containing the FAT filesystem.
    ///
    /// # Returns
    ///
    /// A Result containing a new `Fat` instance if initialization succeeds, or an `std::io::Error` if an error occurs.
    ///
    /// # Errors
    ///
    /// This method may return an error if:
    /// * The partition type is unsupported for FAT filesystems.
    /// * Reading the VBR sector fails.
    /// * Conversion of sector count or other metadata fields fails.
    pub fn load(disk: &mut Disk, partition: &Partition) -> Result<Self, std::io::Error> {
        if disk.get_image_type() == &VhdDisktype::NonVHD {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Not a VHD disk image. Unable to proceed.",
            ));
        }
        let disk_geometry = disk.get_geometry()?;
        let fat_type = match partition.get_partition_type() {
            PartitionType::Fat12MBR => FatType::Fat12,
            PartitionType::Fat16MBR => FatType::Fat16,
            PartitionType::Fat16B => FatType::Fat16B,
            _ => {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "Unsupported partition type for FAT filesystem.",
                ));
            }
        };

        // We don't know the OS yet. [TODO] Implement detection logic for that. Assume IBMDOS200 for now.
        let os = OperatingSystem::IBMDOS200;

        let vbr_address = partition.get_start_sector();
        let vbr_lba = vbr_address.to_lba(&disk_geometry)?;
        let vbr_bytes = disk.read_sector(vbr_lba)?;
        let mut arena = Arena::new();
        let root_directory =
            arena.insert_node(Node::new_directory("rootdir", Vec::new().as_slice()));
        if vbr_bytes[0] == 0 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "No filesystem detected, aborting.",
            ));
        }

        let sector_count = u16::from_le_bytes([vbr_bytes[0x013], vbr_bytes[0x014]]);
        let sectors_per_cluster = vbr_bytes[0x00d];

        let new_fat = Fat {
            fat_type,
            allocation_table: AllocationTable::new(
                sector_count as u32 / sectors_per_cluster as u32,
            ),
            oem_name: Self::oem_name_from_bytes(&vbr_bytes),
            arena,
            root_directory,
            bytes_per_sector: 512,
            sectors_per_cluster,
            reserved_sector_count: u16::from_le_bytes([vbr_bytes[0x00e], vbr_bytes[0x00f]]),
            fat_copies: vbr_bytes[0x010],
            root_dir_entries: u16::from_le_bytes([vbr_bytes[0x011], vbr_bytes[0x012]]),
            sector_count,
            media_descriptor: MediaDescriptor::from_byte(vbr_bytes[0x015]),
            sectors_per_fat: u16::from_le_bytes([vbr_bytes[0x016], vbr_bytes[0x017]]),
            sectors_per_track: u16::from_le_bytes([vbr_bytes[0x018], vbr_bytes[0x019]]),
            heads: u16::from_le_bytes([vbr_bytes[0x01a], vbr_bytes[0x01b]]),
            hidden_sectors: u16::from_le_bytes([vbr_bytes[0x01c], vbr_bytes[0x01d]]),
            bootstrap: os.get_mbr_bootcode(),
            signature: [vbr_bytes[510], vbr_bytes[511]],
            os: None,
        };
        Ok(new_fat)
    }

    /// Compose the Volume Boot Record (VBR) boot sector.
    ///
    /// Constructs the boot sector of the Volume Boot Record (VBR) for the FAT filesystem,
    /// incorporating the provided operating system's VBR jump code, OEM name, BIOS parameter block (BPB),
    /// and VBR boot code.
    ///
    /// # Arguments
    ///
    /// * `os` - A reference to the `OperatingSystem` enum representing the operating system associated with the filesystem.
    ///
    /// # Returns
    ///
    /// A `[u8; 512]` array representing the composed boot sector of the VBR.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::{Fat, OperatingSystem};
    ///
    /// let fat = Fat::new();
    /// let os = OperatingSystem::DOS;
    /// let boot_sector = fat.get_bootsector(&os);
    /// ```
    pub fn get_bootsector(&self, os: &OperatingSystem) -> [u8; 512] {
        let mut bootsector = [0u8; 512];
        let vbr_bootcode = os.get_vbr_bootcode();
        let bpb = self.bpb_dos200(); // Assuming this method exists in your implementation

        // Copy jump code, OEM name, BPB, and VBR boot code to bootsector array
        bootsector[0..3].copy_from_slice(&os.get_vbr_jumpcode());
        bootsector[3..9].copy_from_slice(self.oem_name.as_bytes());
        bootsector[9..23].copy_from_slice(&bpb);
        bootsector[23..(23 + vbr_bootcode.len())].copy_from_slice(&vbr_bootcode);

        // Set boot sector signature bytes
        bootsector[510] = 0x55;
        bootsector[511] = 0xAA;

        bootsector
    }

    /// Create a new file, give the data as a byte slice.
    pub fn mkfile(
        &mut self,
        parent_id: Option<NodeId>,
        filename: &str,
        sysfile: bool,
        data: &[u8],
    ) -> Result<(), std::io::Error> {
        // Determine the DirEntryType based on sysfile flag
        let entry_type = if sysfile {
            DirEntryType::SysFile
        } else {
            DirEntryType::File
        };

        // Attempt to allocate clusters and handle errors appropriately
        let allocated_clusters = self.allocate_clusters(data)?;

        // Write the proper values to the allocation table
        if let Some((last, elements)) = allocated_clusters.split_last() {
            for (i, &current_cluster) in elements.iter().enumerate() {
                let next_cluster = allocated_clusters[i + 1];
                self.allocation_table
                    .allocate_cluster(current_cluster, next_cluster)?;
            }
            // Handle the last cluster, which points to the end of the chain
            self.allocation_table.mark_end_of_chain(*last)?;
        }

        // Insert the new file into the arena and get its NodeId
        let node_id = match entry_type {
            DirEntryType::Directory => self
                .arena
                .insert_node(Node::new_directory(filename, allocated_clusters.as_slice())),
            DirEntryType::File => self.arena.insert_node(Node::new_file(
                filename,
                None,
                allocated_clusters,
                data.len() as u32,
                data.to_vec(),
                false,
            )?),
            DirEntryType::SysFile => self.arena.insert_node(Node::new_file(
                filename,
                None,
                allocated_clusters,
                data.len() as u32,
                data.to_vec(),
                true,
            )?),
        };

        // Add the new entry to the parent directory or use the root directory
        if let Some(pid) = parent_id {
            if let Some(parent_node) = self.arena.get_node_mut(pid) {
                parent_node.value.children.push(node_id);
            } else {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::NotFound,
                    "Parent node not found",
                ));
            }
        } else 
            if let Some(root) = self.arena
                .get_node_mut(self.root_directory) { root.value.children.push(node_id) }
        

        Ok(())
    }

    /// Equivalent of the MS-DOS "MKDIR" command. Provide a name, the system will do the rest.
    pub fn mkdir(
        &mut self,
        parent: Option<NodeId>,
        dir_name: &str,
    ) -> Result<NodeId, std::io::Error> {
        // Set the parent to the root dir if none is given
        let new_start_cluster = self.allocate_cluster()?;
        self.allocation_table.mark_end_of_chain(new_start_cluster)?;
        let allocated_clusters: Vec<u32> = vec![new_start_cluster];

        let parent_entry = parent.unwrap_or(self.root_directory);
        let parent_clusters = if let Some(parent_node) = self.arena.get_node(parent_entry) {
            parent_node.value.get_allocated_clusters().clone()
        } else {
            vec![2] // Default to empty if parent isn't found, though this should be handled more robustly
        };

        // Instantiate the new entry.
        let new_dir_node = Node::new_directory(dir_name, allocated_clusters.as_slice());
        let new_dir_id = self.arena.insert_node(new_dir_node);
        self.arena.add_child(parent_entry, new_dir_id);

        let self_dot_node = Node::new_directory(".", allocated_clusters.as_slice());
        let self_dot_id = self.arena.insert_node(self_dot_node);
        let parent_dot_node = Node::new_directory("..", parent_clusters.as_slice());
        let parent_dot_id = self.arena.insert_node(parent_dot_node);
        self.arena.add_child(new_dir_id, self_dot_id);
        self.arena.add_child(new_dir_id, parent_dot_id);

        Ok(new_dir_id)
    }

    /// Generate a BIOS Parameter Block (BPB) compatible with DOS 2.00.
    ///
    /// Constructs a BPB suitable for DOS 2.00 based on the filesystem parameters.
    /// The BPB format is as follows:
    /// 0-1: Bytes per sector (little-endian)
    /// 2: Sectors per cluster
    /// 3-4: Reserved sector count (little-endian)
    /// 5: Number of FAT copies
    /// 6-7: Maximum number of root directory entries (little-endian)
    /// 8-9: Total sector count (little-endian)
    /// 10: Media descriptor byte
    /// 11-12: Sectors per FAT (little-endian)
    ///
    /// # Returns
    ///
    /// A `[u8; 13]` array representing the BIOS Parameter Block (BPB) compatible with DOS 2.00.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::Fat;
    ///
    /// let fat = Fat::new();
    /// let bpb = fat.bpb_dos200();
    /// ```
    pub fn bpb_dos200(&self) -> [u8; 13] {
        let mut bpb = [0u8; 13];
        bpb[0..2].copy_from_slice(&self.bytes_per_sector.to_le_bytes());
        bpb[2] = self.sectors_per_cluster;
        bpb[3..5].copy_from_slice(&self.reserved_sector_count.to_le_bytes());
        bpb[5] = self.fat_copies;
        bpb[6..8].copy_from_slice(&self.root_dir_entries.to_le_bytes());
        bpb[8..10].copy_from_slice(&self.sector_count.to_le_bytes());
        bpb[10] = self.media_descriptor.as_byte();
        bpb[11..13].copy_from_slice(&self.sectors_per_fat.to_le_bytes());
        bpb
    }

    /// OEM Name string from on-disk bytes
    pub fn oem_name_from_bytes(vbr_bytes: &[u8; 512]) -> String {
        let oem_name_bytes = &vbr_bytes[3..11];
        let mut string = String::from_utf8_lossy(oem_name_bytes).to_string();
        string.retain(|c| c.is_ascii() && c.is_ascii_alphabetic());
        string
    }

    /// Allocate a number of free clusters from the allocation table for a given file.
    ///
    /// This function allocates clusters from the allocation table to accommodate the specified file size.
    /// It calculates the required number of clusters based on the file size and the cluster size,
    /// then allocates them from the allocation table.
    ///
    /// # Arguments
    ///
    /// * `file` - A slice containing the file data.
    ///
    /// # Returns
    ///
    /// A `Result` containing a `Vec` of allocated cluster indices if successful, or an `std::io::Error` if allocation fails.
    ///
    /// # Errors
    ///
    /// This function may return an `std::io::Error` if there are not enough free clusters in the allocation table.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::Fat;
    ///
    /// let mut fat = Fat::new();
    /// let file_data = vec![0u8; 1024]; // Example file data
    /// let allocated_clusters = fat.allocate_clusters(&file_data);
    /// assert!(allocated_clusters.is_ok());
    /// ```
    fn allocate_clusters(&mut self, file: &[u8]) -> Result<Vec<u32>, std::io::Error> {
        let filesize = file.len() as u32;
        let cluster_size = self.bytes_per_sector as u32 * self.sectors_per_cluster as u32;
        let required_cluster_count = Self::round_up_filesize(filesize, cluster_size) / cluster_size; // Require additional cluster for end of chain marker.

        let mut allocated_clusters = Vec::with_capacity(required_cluster_count as usize);
        for _ in 0..required_cluster_count {
            if let Ok(free_cluster) = self.allocate_cluster() {
                allocated_clusters.push(free_cluster);
            } else {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "No free cluster available. File system is full.",
                ));
            }
        }
        Ok(allocated_clusters)
    }

    /// Allocate a single free cluster for things like directory entries.
    fn allocate_cluster(&mut self) -> Result<u32, std::io::Error> {
        if let Some(free_cluster) = self.allocation_table.get_free_cluster() {
            Ok(free_cluster)
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "No free cluster available. File system is full.",
            ))
        }
    }

    /// Calculate the number of sectors per cluster based on total sector count.
    ///
    /// This function calculates the number of sectors per cluster based on the total sector count
    /// and optionally the operating system. For MS/PC-DOS 2.00, it returns a hard-coded value of 8.
    /// For other operating systems, it follows the specification's recommendations.
    ///
    /// # Arguments
    ///
    /// * `sector_count` - Total number of sectors in the filesystem.
    /// * `os` - Optional operating system information.
    ///
    /// # Returns
    ///
    /// The number of sectors per cluster if successful, or an `std::io::Error` if an error occurs.
    ///
    /// # Errors
    ///
    /// This function may return an `std::io::Error` if there are issues determining the sectors per cluster.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::{calculate_sectors_per_cluster, OperatingSystem};
    ///
    /// let sector_count = 65536; // Example sector count
    /// let os = Some(OperatingSystem::IBMDOS200); // Example operating system
    /// let sectors_per_cluster = calculate_sectors_per_cluster(sector_count, os);
    /// assert_eq!(sectors_per_cluster.unwrap(), 8);
    /// ```
    fn calculate_sectors_per_cluster(
        sector_count: u32,
        os: Option<OperatingSystem>,
    ) -> Result<u8, std::io::Error> {
        // Constants for sector counts
        const SMALL_THRESHOLD: u32 = 32768;
        const MEDIUM_THRESHOLD: u32 = 131072;
        const LARGE_THRESHOLD: u32 = 262144;
        const XLARGE_THRESHOLD: u32 = 524288;
        const XXLARGE_THRESHOLD: u32 = 1048576;

        // MS/PC-DOS 2.00 has a hard-coded 8 sectors per cluster.
        if let Some(OperatingSystem::IBMDOS200) = os {
            return Ok(8);
        }

        // Follow the spec for the default case
        match sector_count {
            count if count < SMALL_THRESHOLD => Ok(8),
            count if count < MEDIUM_THRESHOLD => Ok(4),
            count if count < LARGE_THRESHOLD => Ok(8),
            count if count < XLARGE_THRESHOLD => Ok(16),
            count if count < XXLARGE_THRESHOLD => Ok(32),
            _ => Ok(4),
        }
    }

    /// Calculates the number of sectors per FAT copy based on the total sector count,
    /// the number of sectors per cluster, and the type of FAT.
    ///
    /// # Arguments
    ///
    /// * `sector_count` - The total number of sectors in the filesystem.
    /// * `sectors_per_cluster` - The number of sectors per cluster in the filesystem.
    /// * `fat_type` - The type of FAT used in the filesystem (FAT12, FAT16, or FAT16B).
    ///
    /// # Returns
    ///
    /// The number of sectors per FAT copy as a `u16`.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::{calculate_sectors_per_fat, FatType};
    ///
    /// let sector_count = 1000;
    /// let sectors_per_cluster = 8;
    /// let fat_type = FatType::Fat16;
    ///
    /// let sectors_per_fat = calculate_sectors_per_fat(sector_count, sectors_per_cluster, fat_type);
    ///
    /// assert_eq!(sectors_per_fat, 16);
    /// ```
    pub fn calculate_sectors_per_fat(
        sector_count: u32,
        sectors_per_cluster: u32,
        fat_type: FatType,
    ) -> u16 {
        let bytes_per_entry = match fat_type {
            FatType::Fat12 => 1.5,
            FatType::Fat16 | FatType::Fat16B => 2.0,
        };

        let cluster_count = sector_count / sectors_per_cluster;
        let bytes_per_fat: f32 = (bytes_per_entry as f32 * cluster_count as f32) / 512_f32;

        let max_u16 = u16::MAX as f32;
        (bytes_per_fat.min(max_u16)).ceil() as u16
    }

    /// Round file_size up to the nearest multiple of cluster_size to make it fit FAT.
    fn round_up_filesize(file_size: u32, mut cluster_size: u32) -> u32 {
        // Prevent division by zero: 512 is the absolute minimum size.
        if cluster_size == 0 {
            cluster_size = 512;
        }
        let division_result = file_size / cluster_size;
        let remainder = file_size % cluster_size;
        if remainder != 0 {
            (division_result + 1) * cluster_size
        } else {
            file_size
        }
    }

    /// Get the sectors per cluster value
    pub fn get_sectors_per_cluster(&self) -> usize {
        self.sectors_per_cluster as usize
    }

    /// Return the sector address of the first data sector
    pub fn get_first_data_sector(&self) -> u32 {
        // Iterate over data in chunks and write sectors
        let rootdir_sector_count = ((self.root_dir_entries * 32) / 512) & !512;
        let first_data_sector = self.reserved_sector_count
            + (self.fat_copies as u16 * self.sectors_per_fat)
            + rootdir_sector_count;
        first_data_sector as u32
    }

    // Gets children of a directory entry by NodeId and returns references to their DirEntry
    pub fn get_directory_entries(&self, directory_id: NodeId) -> Vec<&DirEntry> {
        // Attempt to get the directory entry from the arena
        if let Some(dir_entry) = self.arena.get_node(directory_id) {
            // Map over the NodeIds in the children vector to fetch the actual DirEntry references
            dir_entry
                .value
                .children
                .iter()
                .filter_map(|&child_id| self.arena.get_node(child_id).map(|node| &node.value))
                .collect()
        } else {
            // If the directory node is not found, return an empty vector
            vec![]
        }
    }

    /// Retrieves the allocated clusters for a directory entry identified by `node_id`.
    pub fn get_allocated_clusters(&self, node_id: NodeId) -> Option<&Vec<u32>> {
        // Access the node from the arena using the node ID
        self.arena
            .get_node(node_id)
            // If the node exists, retrieve the allocated clusters
            .map(|node| node.value.get_allocated_clusters())
    }
}

/// Enum representing different types of FAT (File Allocation Table) file systems.
#[derive(Copy, Clone, PartialEq, Debug)]
pub enum FatType {
    Fat12,
    Fat16,
    Fat16B,
}

impl fmt::Display for FatType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        // Iterate over each partition in the partition table
        match self {
            Self::Fat12 => write!(f, "FAT12"),
            Self::Fat16 => write!(f, "FAT16 original"),
            Self::Fat16B => write!(f, "FAT16 BIG DOS"),
        }
    }
}

/// Enum representing different media descriptors, indicating the type of storage media.
#[derive(Clone, PartialEq, Debug)]
pub enum MediaDescriptor {
    /// 5.25" 320 KB floppy disk formatted with DOS 1.1.
    Floppy320DOS11,
    /// 5.25" 160 KB floppy disk formatted with DOS 1.0.
    Floppy160DOS10,
    /// 5.25" 360 KB floppy disk formatted with DOS 2.0.
    Floppy360DOS20,
    /// 5.25" 180 KB floppy disk formatted with DOS 2.0.
    Floppy180DOS20,
    /// 5.25" 640 KB floppy disk formatted with DOS 2.0.
    Floppy640DOS20,
    /// 3.5" 320 KB floppy disk formatted with DOS 2.0.
    Floppy320DOS20,
    /// 3.5" 1.2 MB floppy disk formatted with DOS 3.0.
    Floppy1200DOS30,
    /// 3.5" 720 KB floppy disk formatted with DOS 3.2.
    Floppy720DOS32,
    /// 3.5" 1.44 MB floppy disk formatted with DOS 3.3.
    Floppy1440DOS33,
    /// 3.5" 2.88 MB floppy disk formatted with DOS 3.3.
    Floppy2880DOS33,
    /// RAM disk.
    RAMDISK,
    /// Hard drive.
    HardDrive,
    /// Unknown or unspecified media descriptor.
    Unknown,
}

impl fmt::Display for MediaDescriptor {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        // Iterate over each partition in the partition table
        match self {
            Self::Floppy320DOS11 => write!(f, "DOS 1.1 floppy 320KB"),
            Self::Floppy160DOS10 => write!(f, "DOS 1.0 floppy 160KB"),
            Self::Floppy360DOS20 => write!(f, "DOS 2.0 floppy 360KB"),
            Self::Floppy180DOS20 => write!(f, "DOS 2.0 floppy 180KB"),
            Self::Floppy640DOS20 => write!(f, "DOS 2.0 floppy 640KB"),
            Self::Floppy320DOS20 => write!(f, "DOS 2.0 floppy 320KB"),
            Self::Floppy1200DOS30 => write!(f, "DOS 3.0 floppy 1.2MB"),
            Self::Floppy720DOS32 => write!(f, "DOS 3.2 floppy 720KB"),
            Self::Floppy1440DOS33 => write!(f, "DOS 3.3 floppy 1.44MB"),
            Self::Floppy2880DOS33 => write!(f, "DOS 3.3 floppy 2.88MB"),
            Self::RAMDISK => write!(f, "RAM drive"),
            Self::HardDrive => write!(f, "Fixed disk"),
            Self::Unknown => write!(f, "Unknown media"),
        }
    }
}

impl MediaDescriptor {
    /// Convert the media descriptor to an on-disk byte. There is duplication here so reversing this isn't straightforward!
    pub fn as_byte(&self) -> u8 {
        match self {
            Self::Floppy320DOS11 => 0xff,
            Self::Floppy160DOS10 => 0xfe,
            Self::Floppy360DOS20 => 0xfd,
            Self::Floppy180DOS20 => 0xfc,
            Self::Floppy640DOS20 => 0xfb,
            Self::Floppy320DOS20 => 0xfa,
            Self::Floppy1200DOS30 => 0xf9,
            Self::Floppy720DOS32 => 0xfa,
            Self::Floppy1440DOS33 => 0xf0,
            Self::Floppy2880DOS33 => 0xf0,
            Self::RAMDISK => 0xfa,
            Self::HardDrive => 0xf8,
            Self::Unknown => 0x00,
        }
    }

    /// Only recognize HardDrive descriptor byte for now. DOSContainer doesn't do floppies (yet).
    pub fn from_byte(byte: u8) -> Self {
        match byte {
            0xf8 => Self::HardDrive,
            _ => Self::Unknown,
        }
    }
}
#[cfg(test)]
mod tests {

    use super::*;

    #[test]
    fn test_oem_name_from_bytes() {
        // Test with valid OEM name bytes
        let mut vbr_bytes_valid: [u8; 512] = [0; 512];
        vbr_bytes_valid[3..12].copy_from_slice(b"MS-DOS5.0");
        assert_eq!(
            Fat::oem_name_from_bytes(&vbr_bytes_valid),
            "MSDOS".to_string()
        );

        // Test with invalid OEM name bytes (non-ASCII characters)
        let mut vbr_bytes_invalid: [u8; 512] = [0; 512];
        vbr_bytes_invalid[3..12].copy_from_slice(b"!@#$%^&*(");
        assert_eq!(Fat::oem_name_from_bytes(&vbr_bytes_invalid), "".to_string());
        // Should return an empty string
    }

    #[test]
    fn test_calculate_sectors_per_fat_fat12() {
        let sector_count = 4096; // FAT12 volume size limit
        let sectors_per_cluster = 1;
        let fat_type = FatType::Fat12;

        let result = Fat::calculate_sectors_per_fat(sector_count, sectors_per_cluster, fat_type);
        assert_eq!(result, 12);
    }

    #[test]
    fn test_calculate_sectors_per_fat_fat16() {
        let sector_count = 60000; // Just above FAT16 volume size limit
        let sectors_per_cluster = 8;
        let fat_type = FatType::Fat16;

        let result = Fat::calculate_sectors_per_fat(sector_count, sectors_per_cluster, fat_type);
        assert_eq!(result, 30);
    }

    #[test]
    fn test_calculate_sectors_per_fat_fat16b() {
        let sector_count = 60000; // Just above FAT16 volume size limit
        let sectors_per_cluster = 4;
        let fat_type = FatType::Fat16B;

        let result = Fat::calculate_sectors_per_fat(sector_count, sectors_per_cluster, fat_type);
        assert_eq!(result, 59);
    }

    #[test]
    fn test_calculate_sectors_per_fat_rounding() {
        let sector_count = 60000;
        let sectors_per_cluster = 7; // Non-integer sectors per cluster
        let fat_type = FatType::Fat16;

        let result = Fat::calculate_sectors_per_fat(sector_count, sectors_per_cluster, fat_type);
        assert_eq!(result, 34);
    }
}
