use super::{cluster::Cluster, cluster::ClusterState, FatType};
use std::collections::BTreeMap;

#[derive(Clone, Debug, PartialEq)]
pub struct AllocationTable {
    //clusters: HashMap<u32, Cluster>,
    clusters: BTreeMap<u32, Cluster>,
    cluster_count: u32,
}

impl AllocationTable {
    pub fn new(cluster_count: u32) -> Self {
        let mut clusters = BTreeMap::new();
        clusters.insert(0, Cluster::new(Some(0xFF8), Some(ClusterState::Allocated)));
        clusters.insert(1, Cluster::new(Some(0xFFF), Some(ClusterState::EndOfChain)));
        AllocationTable {
            clusters,
            cluster_count,
        }
    }

    /// Find the index of the next free cluster in the FAT. Returns None
    /// if the filesystem does not have any more free clusters. Marks any
    /// cluster it returns as Reserved so it can be used by subsequent
    /// operations.
    pub fn get_free_cluster(&mut self) -> Option<u32> {
        // Iterate over existing clusters
        for index in 0..self.cluster_count {
            if let Some(cluster) = self.clusters.get(&index) {
                if cluster.get_state() == ClusterState::Free {
                    self.reserve_cluster(index).unwrap();
                }
            } else {
                // If the cluster doesn't exist in the HashMap, create one that is already reserved.
                self.clusters
                    .insert(index, Cluster::new(None, Some(ClusterState::Reserved)));
                return Some(index);
            }
        }
        None // No free cluster found
    }

    /// Scan the entire file system for clusters marked as reserved and
    /// free them. Useful to do this when filesystem writes didn't go as planned.
    pub fn prune_reserved_clusters(&mut self) -> Result<(), std::io::Error> {
        for index in 0..self.cluster_count {
            if let Some(cluster) = self.clusters.get_mut(&index) {
                if cluster.get_state() == ClusterState::Reserved {
                    cluster.free()?;
                }
            }
        }
        Ok(())
    }

    /// Mark the given cluster as Reserved in preparation for allocation.
    pub fn reserve_cluster(&mut self, index: u32) -> Result<(), std::io::Error> {
        if index > self.cluster_count {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Attempted to reserve a cluster outside filesystem bounds.",
            ));
        }
        // We already have a cluster, reserve it.
        if let Some(cluster) = self.clusters.get_mut(&index) {
            if cluster.get_state() != ClusterState::Free {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "Attempted to reserve an unavailable cluster.",
                ));
            }
            cluster.set_reserved();
            Ok(())
        }
        // Nothing lives at index yet, place a reserved cluster there.
        else {
            self.clusters
                .insert(index, Cluster::new(None, Some(ClusterState::Reserved)));
            Ok(())
        }
    }

    /// Allocate a cluster on the allocation table. A cluster can only ever be allocated if it
    /// is first marked as Reserved in memory. Any state other than Reserved leads to an error.
    pub fn allocate_cluster(&mut self, index: u32, value: u32) -> Result<(), std::io::Error> {
        for (cluster_index, cluster) in self.clusters.iter_mut() {
            if *cluster_index == index {
                if cluster.get_state() != ClusterState::Reserved {
                    return Err(std::io::Error::new(
                        std::io::ErrorKind::InvalidInput,
                        "Attempted to allocate an unavailable cluster.",
                    ));
                }
                cluster.allocate(value)?;
                return Ok(());
            }
        }

        // If the cluster is not found, insert a new one
        self.clusters.insert(
            index,
            Cluster::new(Some(value), Some(ClusterState::Allocated)),
        );

        Ok(())
    }

    /// Mark end of chain on cluster
    pub fn mark_end_of_chain(&mut self, index: u32) -> Result<(), std::io::Error> {
        if index > self.cluster_count {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Attempted to mark end of chain outside bounds of file system.",
            ));
        }
        if let Some(cluster) = self.clusters.get_mut(&index) {
            cluster.set_end_of_chain();
        } else {
            self.clusters
                .insert(index, Cluster::new(None, Some(ClusterState::EndOfChain)));
        }
        Ok(())
    }

    /// Read cluster status from the allocation table
    pub fn get_cluster(&self, cluster_address: u32) -> Result<Cluster, std::io::Error> {
        if cluster_address as usize > self.cluster_count as usize {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Attempted to read cluster outside the partition's boundary.",
            ));
        }
        if let Some(cluster) = self.clusters.get(&cluster_address) {
            Ok(*cluster)
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Fudgies!!",
            ))
        }
    }

    /// Return the number of clusters in the table.
    pub fn len(&self) -> u32 {
        self.clusters.len() as u32
    }

    /// Clippy wants an is_empty method so we give the struct one.
    pub fn is_empty(&self) -> bool {
        matches!(self.clusters.len(), 0)
    }

    /// Convert the struct into the on-disk representation for writing to storage.
    /// How this is done depends on the FAT type.
    pub fn as_bytes(&mut self, fat_type: FatType) -> Result<Vec<u8>, std::io::Error> {
        match fat_type {
            FatType::Fat12 => Ok(self.as_fat12_bytes()?), // FAT12 has a potential error condition, handle it.
            FatType::Fat16 => Ok(self.as_fat16_bytes()?),
            _ => Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Requested FAT type not implemented yet. Please contact developer, this is a todo!",
            )),
        }
    }

    fn as_fat16_bytes(&mut self) -> Result<Vec<u8>, std::io::Error> {
        let mut bytes = Vec::new();

        let max_index = self.clusters.keys().max().copied().unwrap_or(0);

        // Iterate over the cluster indices up to the maximum index
        for index in 0..max_index {
            let mut cluster = self.clusters
                .get(&index)
                .copied()
                .unwrap_or(Cluster::new(Some(0), Some(ClusterState::Free)));

            // Catch end-of-chain markers just before packing into bytes so we can set the proper value.
            if cluster.get_state() == ClusterState::EndOfChain {
                cluster.set_value(0xFFFF);
            }

            let decomposed = (cluster.get_value() as u16).to_le_bytes();
            bytes.push(decomposed[0]);
            bytes.push(decomposed[1]);
        }
        Ok(bytes)
    }

    /// Convert the data structure to the on-disk representation that matches FAT12 for IBM PC-DOS 2.00
    /// From Wikipedia:
    /// The FAT12 file system uses 12 bits per FAT entry, thus two entries span 3 bytes. It is consistently
    /// little-endian: if those three bytes are considered as one little-endian 24-bit number, the 12 least
    /// significant bits represent the first entry (e.g. cluster 0) and the 12 most significant bits the
    /// second (e.g. cluster 1). In other words, while the low eight bits of the first cluster in the row
    /// are stored in the first byte, the top four bits are stored in the low nibble of the second byte,
    /// whereas the low four bits of the subsequent cluster in the row are stored in the high nibble of the
    /// second byte and its higher eight bits in the third byte.
    fn as_fat12_bytes(&mut self) -> Result<Vec<u8>, std::io::Error> {
        let mut bytes = Vec::new();
    
        let max_index = self.clusters.keys().max().copied().unwrap_or(0);
    
        for index in (0..=max_index).step_by(2) {
            let cluster1 = self.clusters.get(&index).copied().unwrap_or_else(|| Cluster::new(Some(0), Some(ClusterState::Free)));
            let cluster2 = self.clusters.get(&(index + 1)).copied().unwrap_or_else(|| Cluster::new(Some(0), Some(ClusterState::Free)));
    
            let bits_1 = cluster1.get_value() as u16 & 0xFFF;
            let bits_2 = cluster2.get_value() as u16 & 0xFFF;
    
            let full_entry = ((bits_2 as u32) << 12) | (bits_1 as u32);
            bytes.extend_from_slice(&full_entry.to_le_bytes()[..3]); // Take only the first 3 bytes
        }
    
        Ok(bytes)
    }
    
}

impl Default for AllocationTable {
    fn default() -> Self {
        Self::new(128)
    }
}