#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Attributes {
    pub archive: bool,
    pub hidden: bool,
    pub system: bool,
    pub read_only: bool,
    pub volume_label: bool,
    pub subdir: bool,
    pub device: bool,
}

impl Attributes {
    pub fn new(
        archive: bool,
        hidden: bool,
        system: bool,
        read_only: bool,
        volume_label: bool,
        subdir: bool,
        device: bool,
    ) -> Self {
        Attributes {
            archive,
            hidden,
            system,
            read_only,
            volume_label,
            subdir,
            device,
        }
    }

    /// Converts the struct to a single byte representing the file attributes.
    ///
    /// Each field of the struct corresponds to a bit in the resulting byte:
    /// - Bit 0: Read-only
    /// - Bit 1: Hidden
    /// - Bit 2: System
    /// - Bit 3: Volume label
    /// - Bit 4: Subdirectory
    /// - Bit 5: Archive
    /// - Bit 6: Device
    ///
    /// Returns:
    /// - `u8`: A byte where each bit represents the state of the corresponding field.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Attributes;
    ///
    /// let attributes = Attributes {
    ///     read_only: true,
    ///     hidden: false,
    ///     system: false,
    ///     volume_label: false,
    ///     subdir: false,
    ///     archive: false,
    ///     device: false,
    /// };
    ///
    /// assert_eq!(attributes.as_byte(), 0b0000_0001);
    /// ```
    pub fn as_byte(&self) -> u8 {
        let mut result: u8 = 0;
        if self.read_only {
            result |= 1 << 0;
        }
        if self.hidden {
            result |= 1 << 1;
        }
        if self.system {
            result |= 1 << 2;
        }
        if self.volume_label {
            result |= 1 << 3;
        }
        if self.subdir {
            result |= 1 << 4;
        }
        if self.archive {
            result |= 1 << 5;
        }
        if self.device {
            result |= 1 << 6;
        }
        result
    }

    /// Generates a new `Attributes` struct from a byte value.
    ///
    /// Each bit of the byte corresponds to a field in the `Attributes` struct:
    /// - Bit 0: Read-only
    /// - Bit 1: Hidden
    /// - Bit 2: System
    /// - Bit 3: Volume label
    /// - Bit 4: Subdirectory
    /// - Bit 5: Archive
    /// - Bit 6: Device
    ///
    /// # Arguments
    ///
    /// * `byte`: The byte representing the file attributes.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_crate_name::Attributes;
    ///
    /// let byte_value: u8 = 0b0000_0010; // Hidden is set
    /// let attributes = Attributes::from_byte(byte_value);
    ///
    /// assert_eq!(attributes.hidden, true);
    /// assert_eq!(attributes.as_byte(), byte_value);
    /// ```
    pub fn from_byte(byte: u8) -> Self {
        Attributes {
            read_only: byte & (1 << 0) != 0,
            hidden: byte & (1 << 1) != 0,
            system: byte & (1 << 2) != 0,
            volume_label: byte & (1 << 3) != 0,
            subdir: byte & (1 << 4) != 0,
            archive: byte & (1 << 5) != 0,
            device: byte & (1 << 6) != 0,
        }
    }
}

impl Default for Attributes {
    fn default() -> Self {
        Attributes {
            archive: true,
            hidden: false,
            system: false,
            read_only: false,
            volume_label: false,
            subdir: false,
            device: false,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_from_to_byte_roundtrip() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: false,
            read_only: true,
            volume_label: false,
            subdir: false,
            device: false,
        };
        let byte = attributes.as_byte();
        let reconstructed = Attributes::from_byte(byte);
        assert_eq!(attributes, reconstructed);
    }

    #[test]
    fn test_as_byte_read_only() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: false,
            read_only: true,
            volume_label: false,
            subdir: false,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0000_0001);
    }

    #[test]
    fn test_as_byte_hidden() {
        let attributes = Attributes {
            archive: false,
            hidden: true,
            system: false,
            read_only: false,
            volume_label: false,
            subdir: false,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0000_0010);
    }

    #[test]
    fn test_as_byte_system() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: true,
            read_only: false,
            volume_label: false,
            subdir: false,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0000_0100);
    }

    #[test]
    fn test_as_byte_volume_label() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: false,
            read_only: false,
            volume_label: true,
            subdir: false,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0000_1000);
    }

    #[test]
    fn test_as_byte_subdir() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: false,
            read_only: false,
            volume_label: false,
            subdir: true,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0001_0000);
    }

    #[test]
    fn test_as_byte_archive() {
        let attributes = Attributes {
            archive: true,
            hidden: false,
            system: false,
            read_only: false,
            volume_label: false,
            subdir: false,
            device: false,
        };
        assert_eq!(attributes.as_byte(), 0b0010_0000);
    }

    #[test]
    fn test_as_byte_device() {
        let attributes = Attributes {
            archive: false,
            hidden: false,
            system: false,
            read_only: false,
            volume_label: false,
            subdir: false,
            device: true,
        };
        assert_eq!(attributes.as_byte(), 0b0100_0000);
    }
}
