#[derive(PartialEq, Debug, Clone, Copy)]
pub enum ClusterState {
    Free,
    Reserved,
    Bad,
    Allocated,
    EndOfChain,
}

#[derive(PartialEq, Debug, Clone, Copy)]
pub struct Cluster {
    value: u32,
    state: ClusterState,
}

impl Cluster {
    /// Constructs a new `Cluster` instance with the specified value and state.
    /// If no value or state is provided, default values are used: 0 for the value and Free for the state.
    ///
    /// # Arguments
    ///
    /// * `value` - An optional u32 value representing the position or index of the cluster.
    /// * `state` - An optional `ClusterState` representing the state of the cluster.
    ///
    /// # Returns
    ///
    /// A new `Cluster` instance initialized with the provided value and state, or with default values if not provided.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::Cluster;
    /// use your_module::ClusterState;
    ///
    /// let cluster1 = Cluster::new(Some(123), Some(ClusterState::Allocated));
    /// let cluster2 = Cluster::new(None, None); // Creates a cluster with default value (0) and state (Free)
    /// ```
    pub fn new(value: Option<u32>, state: Option<ClusterState>) -> Self {
        let default_value = value.unwrap_or(0); // Set default value to 0 if not provided
        let default_state = state.unwrap_or(ClusterState::Free); // Set default state to Free if not provided

        Cluster {
            value: default_value,
            state: default_state,
        }
    }

    pub fn set_value(&mut self, value: u32) {
        self.value = value;
    }

    /// Allocate the cluster
    pub fn allocate(&mut self, value: u32) -> Result<(), std::io::Error> {
        if self.state == ClusterState::Allocated {
            return Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "The current cluster was already allocated. Refusing to proceed.",
            ));
        }
        if self.state == ClusterState::Bad {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Cluster is marked bad. Refusing to allocate a bad cluster.",
            ));
        }
        if self.state == ClusterState::EndOfChain {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Cluster is the end of chain marker for another file. Refusing to allocate cluster.",
            ));
        }
        self.value = value;
        self.state = ClusterState::Allocated;
        Ok(())
    }

    /// Free (deallocate) the cluster
    pub fn free(&mut self) -> Result<(), std::io::Error> {
        if self.state == ClusterState::Bad {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Refusing to deallocate a bad cluster.",
            ));
        }
        self.value = 0;
        self.state = ClusterState::Free;
        Ok(())
    }

    /// Getter method for the cluster value
    pub fn get_value(&self) -> u32 {
        self.value
    }

    /// Getter method for the cluster state
    pub fn get_state(&self) -> ClusterState {
        self.state
    }

    /// Set as end-of-chain marker
    pub fn set_end_of_chain(&mut self) {
        self.value = 0xFFFF;
        self.state = ClusterState::EndOfChain;
    }
    
    /// Setter method for Bad state
    pub fn set_bad(&mut self) {
        self.value = 0;
        self.state = ClusterState::Bad;
    }

    pub fn set_reserved(&mut self) {
        self.value = 1;
        self.state = ClusterState::Reserved;
    }
}
