use super::{attributes::Attributes, filename::FileName, timedate::TimeDate};
use std::{cmp::min, fmt};

pub struct NodeRef<'a, T> {
    node_id: NodeId,
    arena: &'a Arena<T>,
}

impl<'a, T> NodeRef<'a, T> {
    pub fn new(node_id: NodeId, arena: &'a Arena<T>) -> Self {
        NodeRef { node_id, arena }
    }
}

impl<'a> fmt::Display for NodeRef<'a, DirEntry> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(node) = self.arena.get_node(self.node_id) {
            write!(f, "{}", node.value)
        } else {
            write!(f, "Invalid NodeId")
        }
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct DirEntry {
    filename: FileName,
    attributes: Attributes,
    create_timedate: TimeDate,
    pub file_size: u32,
    pub entry_type: DirEntryType,
    pub children: Vec<NodeId>,
    pub allocated_clusters: Vec<u32>,
    data: Vec<u8>,
}

#[derive(Clone, Debug, Eq, PartialEq)]
pub struct Node<T> {
    pub value: T,
    pub parent: Option<NodeId>, // Optionally keep track of the parent
}

impl Node<DirEntry> {
    /// Creates a new file node with given properties.
    pub fn new_file(
        filename: &str,
        create_timedate: Option<TimeDate>,
        allocated_clusters: Vec<u32>,
        file_size: u32,
        data: Vec<u8>,
        sysfile: bool, // Indicates whether the file is a system file or regular file
    ) -> Result<Self, std::io::Error> {
        // Create the file entry using the DirEntry's create method
        let file_entry = DirEntry::create_file(
            filename,
            create_timedate,
            &allocated_clusters,
            file_size,
            data,
            sysfile,
        )?;

        // Return the new Node encapsulating the DirEntry
        Ok(Node {
            value: file_entry,
            parent: None, // Initially, no parent is set
        })
    }
    pub fn new_directory(dir_name: &str, allocated_clusters: &[u32]) -> Self {
        Node {
            value: DirEntry::new(
                dir_name,
                DirEntryType::Directory,
                None,
                &allocated_clusters.to_vec(),
                0,
                None,
            )
            .expect("Boom!"),
            parent: None,
        }
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct Arena<T> {
    nodes: Vec<Node<T>>,
}

impl Default for Arena<DirEntry> {
    fn default() -> Self {
        Self::new()
    }
}

impl Arena<DirEntry> {
    pub fn new() -> Self {
        Arena { nodes: Vec::new() }
    }

    pub fn new_node(&mut self, entry: DirEntry) -> NodeId {
        let node_id = NodeId(self.nodes.len());
        self.nodes.push(Node {
            value: entry,
            parent: None,
        });
        node_id
    }

    pub fn add_child(&mut self, parent_id: NodeId, child_id: NodeId) {
        if let Some(node) = self.nodes.get_mut(parent_id.0) {
            node.value.children.push(child_id);
        }

        if let Some(node) = self.nodes.get_mut(child_id.0) {
            node.parent = Some(parent_id);
        }
    }

    /// Inserts a new node into the arena and returns a NodeId that can be used to reference it.
    pub fn insert_node(&mut self, node: Node<DirEntry>) -> NodeId {
        let node_id = NodeId(self.nodes.len());
        self.nodes.push(node);
        node_id
    }

    /// Retrieves an immutable reference to the node specified by `node_id`.
    /// Returns `None` if no node exists at that index.
    pub fn get_node(&self, node_id: NodeId) -> Option<&Node<DirEntry>> {
        self.nodes.get(node_id.0)
    }

    /// Retrieves a mutable reference to the node specified by `node_id`.
    /// Returns `None` if no node exists at that index.
    pub fn get_node_mut(&mut self, node_id: NodeId) -> Option<&mut Node<DirEntry>> {
        self.nodes.get_mut(node_id.0)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct NodeId(usize);

/// Enum to differentiate between different types of DirEntry. These types streamline
/// the definition of default values.
#[derive(Clone, Copy, Debug, PartialEq)]
pub enum DirEntryType {
    File,
    SysFile,
    Directory,
}

impl fmt::Display for DirEntryType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            DirEntryType::Directory => write!(f, "Subdirectory"),
            DirEntryType::File => write!(f, "Regular file"),
            DirEntryType::SysFile => write!(f, "System file"),
        }
    }
}

impl fmt::Display for DirEntry {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "File name     : {}\n\
             Entry type    : {}\n\
             ",
            self.filename.base_as_string(),
            self.entry_type
        )
    }
}

impl DirEntry {
    /// Public entry point for creating new DirEntry instances. The function itself branches out
    /// depending on the type of entry desired.
    pub fn new(
        filename: &str,
        entry_type: DirEntryType,
        create_timedate: Option<TimeDate>,
        allocated_clusters: &Vec<u32>,
        file_size: u32,
        data: Option<&[u8]>,
    ) -> Result<Self, std::io::Error> {
        match entry_type {
            DirEntryType::File => Self::create_file(
                filename,
                create_timedate,
                allocated_clusters,
                file_size,
                data.unwrap().to_vec(),
                false,
            ),
            DirEntryType::SysFile => Self::create_file(
                filename,
                create_timedate,
                allocated_clusters,
                file_size,
                data.unwrap().to_vec(),
                true,
            ),
            DirEntryType::Directory => {
                Self::create_directory(filename, create_timedate, allocated_clusters)
            }
        }
    }

    /// Create a regular file with Archive set, all other attributes unset.
    /// If you don't provide a TimeDate, the current system time on the host machine will be used.
    fn create_file(
        filename: &str,
        create_timedate: Option<TimeDate>,
        allocated_clusters: &[u32],
        file_size: u32,
        data: Vec<u8>,
        sysfile: bool,
    ) -> Result<Self, std::io::Error> {
        let attributes = match sysfile {
            false => Attributes::new(true, false, false, false, false, false, false),
            true => Attributes::new(true, true, true, true, false, false, false),
        };
        let entry_type = match sysfile {
            true => DirEntryType::SysFile,
            false => DirEntryType::File,
        };
        // Either use the provided TimeDate or get the current system TimeDate
        let effective_timedate = create_timedate.unwrap_or_else(TimeDate::now);

        // Create the file name, handling potential errors
        let file_name_result = FileName::new(filename);
        match file_name_result {
            Ok(file_name) => Ok(DirEntry {
                filename: file_name,
                attributes,
                create_timedate: effective_timedate,
                file_size,
                entry_type,
                children: Vec::new(),
                allocated_clusters: allocated_clusters.to_vec(),
                data,
            }),
            Err(e) => Err(e),
        }
    }

    fn create_directory(
        filename: &str,
        create_timedate: Option<TimeDate>,
        allocated_clusters: &[u32],
    ) -> Result<Self, std::io::Error> {
        // Set of attributes appropriate to a directory
        let attributes = Attributes::new(false, false, false, false, false, true, false);
        // This is the acual directory we're making
        let new_entry = match create_timedate {
            None => DirEntry {
                filename: FileName::new(filename)?,
                attributes,
                create_timedate: TimeDate::now(),
                file_size: 0,
                entry_type: DirEntryType::Directory,
                children: Vec::new(),
                allocated_clusters: allocated_clusters.to_vec(),
                data: Vec::new(),
            },
            _ => DirEntry {
                filename: FileName::new(filename)?,
                attributes,
                create_timedate: create_timedate.unwrap(),
                file_size: 0,
                entry_type: DirEntryType::Directory,
                children: Vec::new(),
                allocated_clusters: allocated_clusters.to_vec(),
                data: Vec::new(),
            },
        };
        Ok(new_entry)
    }

    /// Converts the fields of a `DirEntry` struct into a byte array representing a directory entry.
    ///
    /// This function takes various fields from a `DirEntry` struct representing a directory entry and converts them into
    /// a byte array that replicates the MS-DOS on-disk format.
    ///
    /// # Arguments
    ///
    /// * `self` - A reference to the `DirEntry` struct containing the fields to be converted.
    ///
    /// # Returns
    ///
    /// A Result containing a 32-byte array representing the directory entry if the conversion succeeds,
    /// or an `std::io::Error` if an error occurs during the conversion process.
    ///
    pub fn as_bytes(&self) -> Result<[u8; 32], std::io::Error> {
        let mut result = [0u8; 32];
        result[0..11].copy_from_slice(&self.filename.as_bytes());
        result[11] = self.attributes.as_byte();
        result[22..24].copy_from_slice(&self.create_timedate.creation_time_bytes()?);
        result[24..26].copy_from_slice(&self.create_timedate.creation_date_bytes()?);
        result[26..28].copy_from_slice(&self.start_cluster_as_bytes());
        result[28..32].copy_from_slice(&self.file_size.to_le_bytes());
        Ok(result)
    }

    /// Very rudimentary function to read a direntry from bytes, which in turn can be used
    /// when reading a directory from disk. DOES NOT WORK YET!! [TODO]
    pub fn from_bytes(bytes: &[u8; 32]) -> Result<DirEntry, std::io::Error> {
        let filename_bytes = &bytes[0..11];
        let filename = FileName::from_bytes(filename_bytes)?;
        let mut datetime_bytes = [0u8; 4];
        datetime_bytes[0..2].copy_from_slice(&bytes[2..4]);
        datetime_bytes[2..4].copy_from_slice(&bytes[0..2]);
        let new_direntry = DirEntry {
            filename,
            attributes: Attributes::from_byte(bytes[11]),
            create_timedate: TimeDate::from_bytes(&datetime_bytes)?,
            file_size: 0,
            entry_type: DirEntryType::File, // This is just as wrong as the rest of this function.
            children: Vec::new(),
            allocated_clusters: Vec::new(),
            data: Vec::new(),
        };
        Ok(new_direntry)
    }

    /// Checks if a given string meets MS-DOS requirements.
    ///
    /// The requirements are:
    /// - No longer than 8 characters
    /// - Upper case letters A-Z
    /// - Numbers 0-9
    /// - Space
    /// - Special characters: ! # $ % & ' ( ) - @ ^ _ ` { } ~
    /// - Characters in the ranges 128-228 and 230-255
    /// - Trailing spaces are considered invalid
    ///
    /// # Arguments
    ///
    /// * `filename` - A string slice that represents the filename to be checked.
    ///
    /// # Returns
    ///
    /// * `true` if the filename meets all requirements and has no trailing spaces, otherwise `false`.
    ///
    /// # Examples
    ///
    /// ```
    /// let valid_filename = "EXAMPLE123.TXT";
    /// assert!(is_valid_filename(valid_filename));
    ///
    /// let invalid_filename = "example123.txt";
    /// assert!(!is_valid_filename(invalid_filename));
    /// ```
    pub fn is_valid_filename(filename: &str) -> bool {
        if filename.is_empty() || filename.len() > 8 {
            return false;
        }
        let valid_chars: Vec<char> = (0..=9)
            .map(|d| char::from_digit(d, 10).unwrap())
            .chain('A'..='Z')
            .chain(vec![
                ' ', '!', '#', '$', '%', '&', '\'', '(', ')', '-', '@', '^', '_', '`', '{', '}',
                '~',
            ])
            .chain((128..=228).map(char::from))
            .chain((230..=255).map(char::from))
            .collect();

        let has_trailing_space = filename.ends_with(' ');
        let all_valid_chars = filename.chars().all(|c| valid_chars.contains(&c));

        all_valid_chars && !has_trailing_space
    }

    /// Converts the start cluster number to a little-endian byte representation.
    ///
    /// # Returns
    ///
    /// An array of 2 bytes representing the start cluster number in little-endian format.
    ///
    /// # Example
    ///
    /// ```
    /// use your_module_name::YourStruct;
    ///
    /// let your_struct = YourStruct {
    ///     start_cluster: 1234,
    /// };
    ///
    /// let bytes = your_struct.start_cluster_as_bytes();
    /// assert_eq!(bytes, [0x32, 0x04]); // Assuming the start cluster number is 1234
    /// ```
    pub fn start_cluster_as_bytes(&self) -> [u8; 2] {
        if self.allocated_clusters.is_empty() {
            // Our parent is the root directory
            return [0,0];
        }
        let cluster: u16 = min(self.allocated_clusters[0], 65535) as u16;
        cluster.to_le_bytes()
    }

    /// Converts the file size to a little-endian byte representation.
    ///
    /// # Returns
    ///
    /// An array of 4 bytes representing the file size in little-endian format.
    ///
    /// # Example
    ///
    /// ```
    /// use your_module_name::YourStruct;
    ///
    /// let your_struct = YourStruct {
    ///     file_size: 1024,
    /// };
    ///
    /// let bytes = your_struct.filesize_as_bytes();
    /// assert_eq!(bytes, [0x00, 0x04, 0x00, 0x00]); // Assuming the file size is 1024
    /// ```
    pub fn filesize_as_bytes(&self) -> [u8; 4] {
        self.file_size.to_le_bytes()
    }

    /// The start cluster is either 0, or the first entry in the allocated clusters vec.
    pub fn get_start_cluster(&self) -> u32 {
        if self.allocated_clusters.is_empty() {
            0
        } else {
            self.allocated_clusters[0]
        }
    }

    pub fn get_allocated_clusters(&self) -> &Vec<u32> {
        &self.allocated_clusters
    }

    pub fn get_children(&mut self) -> &Vec<NodeId> {
        &self.children
    }

    /// Adds a child node to this directory entry and updates necessary metadata
    pub fn add_child(
        &mut self,
        child_id: NodeId,
        arena: &mut Arena<DirEntry>,
    ) -> Result<(), std::io::Error> {
        // Add the child NodeId to the children list
        self.children.push(child_id);

        // Attempt to update the data field with the serialized bytes of all children
        match self.children_bytes(arena) {
            Ok(bytes) => {
                self.data = bytes.to_vec();
                Ok(())
            }
            Err(e) => Err(e),
        }
    }

    pub fn count_children(&self) -> usize {
        self.children.len()
    }

    pub fn directory_name(&self) -> String {
        self.filename.base_as_string()
    }

    pub fn get_type(&self) -> DirEntryType {
        self.entry_type
    }

    pub fn get_data(&self) -> &[u8] {
        self.data.as_slice()

    }

    pub fn is_directory(&self) -> bool {
        matches!(self.get_type(), DirEntryType::Directory)
    }

    pub fn children_bytes(&self, arena: &Arena<DirEntry>) -> Result<Vec<u8>, std::io::Error> {
        let mut bytes = Vec::new();
        for child_id in &self.children {
            // Fetch the child DirEntry from the arena using the NodeId
            if let Some(child_node) = arena.get_node(*child_id) {
                // Serialize the child DirEntry
                // `as_bytes` needs to be a method implemented for serializing DirEntry instances
                bytes.extend_from_slice(&child_node.value.as_bytes()?);
            } else {
                // Handle the case where the child node cannot be found in the arena
                return Err(std::io::Error::new(
                    std::io::ErrorKind::NotFound,
                    "Child node not found in the arena",
                ));
            }
        }
        Ok(bytes)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn as_bytes() {
        let creation_time = TimeDate {
            year: 1983,
            month: 04,
            day: 23,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let dir_entry = DirEntry {
            filename: FileName::new("ibmbio.com").unwrap(),
            attributes: Attributes::default(),
            create_timedate: creation_time,
            file_size: 4608,
            entry_type: DirEntryType::File, // Dummy placeholder
            children: Vec::new(),
            allocated_clusters: vec![2, 3, 4],
            data: Vec::new(),
        };
        // Semi-representative entry that reflects IBMBIO.COM on April 4 1983.
        let reference: [u8; 32] = [
            73, 66, 77, 66, 73, 79, 32, 32, 67, 79, 77, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6,
            151, 2, 0, 0, 18, 0, 0,
        ];
        let result = dir_entry.as_bytes().unwrap();
        assert_eq!(result, reference);
    }
}
