/// Abstraction for a FAT12/16/32 filename data structure. This is currently
/// only valid for the so-called "short" or 8.3 filename type. VFAT and FAT32
/// support long names, but those filesystems haven't been implemented on
/// DOSContainer yet.
#[derive(Clone, Debug, PartialEq)]
pub struct FileName {
    filename: String,
    extension: String,
}

impl FileName {
    /// Creates a new `FileName` instance from a given filename.
    ///
    /// The `filename` parameter should be in the MS-DOS 8.3 format,
    /// consisting of up to 8 characters for the filename (base) and
    /// up to 3 characters for the extension separated by a dot.
    ///
    /// # Examples
    ///
    /// ```
    /// use my_module::FileName;
    ///
    /// let file_name_with_extension = FileName::new("verylongfilename.txt");
    /// assert_eq!(file_name_with_extension.filename, "VERYLONG");
    /// assert_eq!(file_name_with_extension.extension, "TXT");
    ///
    /// let file_name_without_extension = FileName::new("shortname");
    /// assert_eq!(file_name_without_extension.filename, "SHORTNAM");
    /// assert_eq!(file_name_without_extension.extension, "");
    /// ```
    pub fn new(filename: &str) -> Result<Self, std::io::Error> {
        // Handle the exceptional cases of 'self' and 'parent' for directories.
        if filename == "." || filename == ".." {
            let base = filename.to_string();
            let ext = String::new();
            return Ok(FileName {
                filename: base,
                extension: ext,
            });
        }
        if let Some(dot_index) = filename.find('.') {
            let (base, extension) = filename.split_at(dot_index);
            let base = base.chars().take(8).collect::<String>(); // Truncate base to 8 characters
            let extension = extension.chars().skip(1).take(3).collect::<String>(); // Skip the dot and truncate extension to 3 characters
            if base.is_empty() {
                Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "Base name is empty after processing",
                ))
            } else {
                Ok(FileName {
                    filename: base.to_uppercase(),
                    extension: extension.to_uppercase(),
                })
            }
        } else {
            let base = filename.chars().take(8).collect::<String>(); // Truncate base to 8 characters
            if base.is_empty() {
                Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidInput,
                    "Base name is empty after processing",
                ))
            } else {
                Ok(FileName {
                    filename: base.to_uppercase(),
                    extension: String::new(),
                })
            }
        }
    }

    /// Attempts to create a `FileName` instance from a byte slice.
    ///
    /// # Arguments
    ///
    /// * `bytes` - A byte slice representing the filename and extension. The byte slice should
    ///   be exactly 11 bytes long, where the first 8 bytes represent the filename and the last
    ///   3 bytes represent the extension.
    ///
    /// # Errors
    ///
    /// Returns an error of type `std::io::Error` if the input byte slice is not exactly 11 bytes
    /// long, or if the bytes do not represent a valid filename or extension (e.g., contain
    /// non-uppercase ASCII characters).
    ///
    /// # Examples
    ///
    /// ```
    /// use my_module::FileName;
    ///
    /// // Valid byte sequence representing filename "HELLO" with extension "TXT"
    /// let bytes = [72, 69, 76, 76, 79, 32, 32, 32, 84, 88, 84];
    /// let file_name = FileName::from_bytes(&bytes).unwrap();
    /// assert_eq!(file_name.filename, "HELLO");
    /// assert_eq!(file_name.extension, "TXT");
    ///
    /// // Invalid byte sequence (not exactly 11 bytes long)
    /// let invalid_bytes = [72, 69, 76, 76, 79, 32, 32, 32, 84, 88];
    /// let result = FileName::from_bytes(&invalid_bytes);
    /// assert!(result.is_err());
    /// ```
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, std::io::Error> {
        if bytes.len() < 11 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidInput,
                "Input bytes too short",
            ));
        }

        // Split the bytes into filename and extension parts
        let filename_bytes = &bytes[..8];
        let extension_bytes = &bytes[8..11];

        // Convert bytes to strings, removing any trailing spaces
        let filename = String::from_utf8_lossy(filename_bytes)
            .trim_end()
            .to_string();
        let extension = String::from_utf8_lossy(extension_bytes)
            .trim_end()
            .to_string();

        Ok(FileName {
            filename,
            extension,
        })
    }

    /// Returns an array of bytes representing the filename and extension.
    ///
    /// The filename is truncated to 8 characters, and the extension is padded to 3 characters with spaces (0x20).
    /// The resulting byte sequence is exactly 11 bytes long.
    ///
    /// # Examples
    ///
    /// ```
    /// use my_module::FileName;
    ///
    /// let file_name = FileName { filename: String::from("verylongfilename"), extension: String::from("txt") };
    /// let bytes = file_name.as_bytes();
    /// assert_eq!(bytes, [86, 69, 82, 89, 76, 79, 78, 71, 32, 32, 32]);
    /// ```
    pub fn as_bytes(&self) -> [u8; 11] {
        // Truncate the filename if it's longer than 8 characters
        let truncated_filename = if self.filename.len() > 8 {
            &self.filename[..8]
        } else {
            &self.filename
        };

        // Pad the filename if it's shorter than 8 characters
        let mut filename_padded = String::from(truncated_filename);
        filename_padded.push_str(&" ".repeat(8 - truncated_filename.len()));

        // Pad the extension if it's shorter than 3 characters
        let mut extension_padded = String::from(&self.extension);
        extension_padded.push_str(&" ".repeat(3 - self.extension.len()));

        // Convert the padded filename and extension to bytes
        let filename_bytes = filename_padded.as_bytes();
        let extension_bytes = extension_padded.as_bytes();

        // Combine the filename and extension bytes into a single byte array
        let mut result = [0u8; 11];
        result[..8].copy_from_slice(filename_bytes);
        result[8..].copy_from_slice(extension_bytes);

        result
    }

    pub fn base_as_string(&self) -> String {
        self.filename.to_owned()
    }

    pub fn ext_as_string(&self) -> String {
        self.extension.to_owned()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_valid_filename_with_extension() {
        let filename = "verylongfilename.txt";
        let file_name = FileName::new(filename).unwrap();
        assert_eq!(file_name.filename, "VERYLONG");
        assert_eq!(file_name.extension, "TXT");
    }

    #[test]
    fn test_valid_filename_without_extension() {
        let filename = "shortname";
        let file_name = FileName::new(filename).unwrap();
        assert_eq!(file_name.filename, "SHORTNAM");
        assert_eq!(file_name.extension, "");
    }

    #[test]
    fn test_empty_filename() {
        let filename = "";
        let file_name = FileName::new(filename);
        assert!(file_name.is_err());
    }

    #[test]
    fn test_too_long_filename() {
        let filename = "thisfilenameistoolong.txt";
        let file_name = FileName::new(filename).unwrap();
        assert_eq!(file_name.filename, "THISFILE");
        assert_eq!(file_name.extension, "TXT");
    }

    #[test]
    fn test_empty_base_name_after_processing() {
        let filename = ".txt";
        let file_name = FileName::new(filename);
        assert!(file_name.is_err());
    }

    #[test]
    fn test_from_bytes_valid() {
        // Test valid input
        let bytes = b"FILENAMEEXT";
        let result = FileName::from_bytes(&bytes[..]).unwrap();
        assert_eq!(result.filename, "FILENAME".to_string());
        assert_eq!(result.extension, "EXT".to_string());
    }

    #[test]
    fn test_from_bytes_invalid_short_input() {
        // Test input too short
        let bytes = b"SHORT";
        let result = FileName::from_bytes(&bytes[..]);
        assert!(result.is_err());
        assert_eq!(
            result.err().unwrap().kind(),
            std::io::ErrorKind::InvalidInput
        );
    }

    #[test]
    fn test_from_bytes_valid_utf8_with_extra_bytes() {
        // Test valid UTF-8 input with extra bytes
        let bytes = b"FILENAMEEXTEXTRA";
        let result = FileName::from_bytes(&bytes[..]).unwrap();
        assert_eq!(result.filename, "FILENAME".to_string());
        assert_eq!(result.extension, "EXT".to_string());
    }

    #[test]
    fn test_bytes_roundtrip() {
        let filename = FileName::new("command.com").unwrap();
        let bytes = filename.as_bytes();
        let reconstituted = FileName::from_bytes(&bytes).unwrap();
        assert_eq!(filename, reconstituted);
    }
}
