use chrono::{Datelike, Local, Timelike};
/// Struct for the FAT TimeDate data structure. This is a very
/// basic implementation. Do not try to do anything stupid like
/// setting the date to February 30th or other invalid things.
/// This struct does not do extensive input validation, just
/// general range checking.
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct TimeDate {
    pub year: u32,
    pub month: u8,
    pub day: u8,
    pub hour: u8,
    pub minute: u8,
    pub second: u8,
}

impl TimeDate {
    pub fn now() -> Self {
        let now = Local::now();

        TimeDate {
            year: now.year() as u32,
            month: now.month() as u8,
            day: now.day() as u8,
            hour: now.hour() as u8,
            minute: now.minute() as u8,
            second: 0_u8,
        }
    }

    /// Encodes the creation date of a file into a 2-byte array representing the date in MS-DOS format.
    ///
    /// # Arguments
    ///
    /// * `self` - A reference to the structure containing the file creation date information.
    ///
    /// # Returns
    ///
    /// A Result containing a 2-byte array representing the date encoded as MS-DOS expects it in a directory entry
    /// or an error if the year is out of range for MS-DOS (1980-2107) or if the month or day are invalid.
    ///
    /// # Examples
    ///
    /// ```
    /// use your_module::TimeDate;
    ///
    /// let time_date = TimeDate {
    ///     year: 2022,
    ///     month: 10,
    ///     day: 15,
    /// };
    ///
    /// let result = time_date.creation_date_bytes();
    /// assert!(result.is_ok());
    /// ```
    pub fn creation_date_bytes(&self) -> Result<[u8; 2], std::io::Error> {
        // Check if year is within MS-DOS date range
        if self.year < 1980 || self.year > 2107 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Year out of range for MS-DOS.",
            ));
        }

        // Check if month and day are valid
        if self.day > 31 || self.month > 12 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Invalid month or day.",
            ));
        }

        // Encode the year (subtract 1980)
        let encoded_year: u16 = (self.year - 1980) as u16;

        // Extract bits for year, month, and day
        let year_bits: u16 = (encoded_year & 0b0111_1111) << 9;
        let month_bits: u16 = (self.month as u16 & 0b0000_1111) << 5;
        let day_bits: u16 = self.day as u16 & 0b0001_1111;

        // Combine the extracted bits into a single u16 value
        let result: u16 = year_bits | month_bits | day_bits;

        // Convert result to big-endian bytes. This is *reallY* weird for now
        // because FAT is little-endian. This is how it is on-disk with IBM though,
        // so this is what we do.
        Ok(result.to_be_bytes())
    }

    /// Parses a `[u8; 2]` array into a `TimeDate` structure.
    ///
    /// # Arguments
    ///
    /// * `bytes`: The bytes representing the date.
    ///
    /// # Returns
    ///
    /// A `Result` containing the parsed date if successful, or an `std::io::Error` if parsing fails.
    ///
    /// # Examples
    ///
    /// ```
    /// use std::io::ErrorKind;
    /// use your_crate_name::TimeDate;
    ///
    /// let bytes: [u8; 2] = [0b0000_0000, 0b0000_0000]; // Date: January 1, 1980
    /// let date = TimeDate::parse_creation_date_bytes(bytes);
    ///
    /// assert_eq!(date.unwrap().year, 1980);
    /// assert_eq!(date.unwrap().month, 1);
    /// assert_eq!(date.unwrap().day, 1);
    /// ```
    pub fn date_from_bytes(bytes: [u8; 2]) -> Result<Self, std::io::Error> {
        // Convert the big-endian bytes to a u16 value
        let result = u16::from_be_bytes(bytes);

        // Extract year, month, and day bits from the u16 value
        let year = ((result >> 9) & 0b0111_1111) as u32 + 1980;
        let month = ((result >> 5) & 0b0000_1111) as u8;
        let day = (result & 0b0001_1111) as u8;

        // Validate month and day
        if day > 31 || month > 12 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Invalid month or day.",
            ));
        }

        Ok(TimeDate {
            year,
            month,
            day,
            // For simplicity, assume default values for hour, minute, and second
            hour: 0,
            minute: 0,
            second: 0,
        })
    }

    pub fn creation_time_bytes(&self) -> Result<[u8; 2], std::io::Error> {
        // Validate the fields before using them
        if self.hour > 23 || self.minute > 59 || self.second > 59 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Invalid value for hour, minute or second.",
            ));
        }
        let hour_bits: u16 = (self.hour as u16 & 0b0001_1111) << 11;
        let minute_bits: u16 = (self.minute as u16 & 0b0011_1111) << 5;
        let second_bits: u16 = (self.second / 2) as u16 & 0b0001_1111; // Divide by 2 for MS-DOS doublesecond datatype
        let result = hour_bits | minute_bits | second_bits;
        Ok(result.to_be_bytes())
    }

    pub fn time_from_bytes(bytes: &[u8; 2]) -> Result<Self, std::io::Error> {
        // Convert the big-endian bytes to a u16 value
        let result = u16::from_be_bytes(*bytes);

        // Extract hour, minute, and second bits from the u16 value
        let hour = ((result >> 11) & 0b0001_1111) as u8;
        let minute = ((result >> 5) & 0b0011_1111) as u8;
        let second = ((result & 0b0001_1111) * 2) as u8; // Multiply by 2 to get back the original second value

        // Round down the second value to the nearest even number
        let second = if second % 2 == 1 { second - 1 } else { second };

        // Validate hour, minute, and second
        if hour > 23 || minute > 59 || second > 59 {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Invalid value for hour, minute, or second.",
            ));
        }

        Ok(TimeDate {
            year: 1984, // For simplicity, assume default values for year, month, and day
            month: 4,
            day: 2,
            hour,
            minute,
            second,
        })
    }

    pub fn from_bytes(bytes: &[u8; 4]) -> Result<TimeDate, std::io::Error> {
        // Convert the big-endian bytes to two u16 values
        let date_result = u16::from_be_bytes([bytes[0], bytes[1]]);
        let time_result = u16::from_be_bytes([bytes[2], bytes[3]]);

        // Extract year, month, and day bits from the date u16 value
        let year = ((date_result >> 9) & 0b0111_1111) as u32 + 1980;
        let month = ((date_result >> 5) & 0b0000_1111) as u8;
        let day = (date_result & 0b0001_1111) as u8;

        // Extract hour, minute, and second bits from the time u16 value
        let hour = ((time_result >> 11) & 0b0001_1111) as u8;
        let minute = ((time_result >> 5) & 0b0011_1111) as u8;
        let second = ((time_result & 0b0001_1111) * 2) as u8; // Multiply by 2 to get back the original second value

        // Round down the second value to the nearest even number
        let second = if second % 2 == 1 { second - 1 } else { second };

        // Validate year, month, day, hour, minute, and second
        if !(1980..=2107).contains(&year)
            || day > 31
            || month > 12
            || hour > 23
            || minute > 59
            || second > 59
        {
            return Err(std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                "Invalid value for year, month, day, hour, minute, or second.",
            ));
        }

        Ok(TimeDate {
            year,
            month,
            day,
            hour,
            minute,
            second,
        })
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_creation_time_valid() {
        let date = TimeDate {
            year: 1983,
            month: 04,
            day: 23,
            hour: 11,
            minute: 43,
            second: 5,
        };
        let bytes = date.creation_time_bytes().unwrap();
        let testvalue: [u8; 2] = [93, 98];
        assert_eq!(bytes, testvalue);
    }

    #[test]
    fn test_creation_time_invalid() {
        let date = TimeDate {
            year: 1983,
            month: 04,
            day: 23,
            hour: 15,
            minute: 70,
            second: 93,
        };
        let result = date.creation_time_bytes();
        assert!(result.is_err());
        assert_eq!(
            result.unwrap_err().to_string(),
            "Invalid value for hour, minute or second."
        );
    }

    #[test]
    fn test_creation_date_roundtrip() {
        let date = TimeDate {
            year: 1983,
            month: 4,
            day: 23,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let date_bytes = date.creation_date_bytes().unwrap();
        let reconstructed = TimeDate::date_from_bytes(date_bytes).unwrap();
        assert_eq!(date, reconstructed);
    }

    #[test]
    fn test_creation_time_roundtrip() {
        let time = TimeDate {
            year: 1984,
            month: 4,
            day: 2,
            hour: 21,
            minute: 35,
            second: 48,
        };
        let time_bytes = time.creation_time_bytes().unwrap();
        let reconstructed = TimeDate::time_from_bytes(&time_bytes).unwrap();
        assert_eq!(time, reconstructed);
    }

    #[test]
    fn test_full_roundtrip() {
        let timedate = TimeDate {
            year: 1984,
            month: 9,
            day: 31,
            hour: 21,
            minute: 42,
            second: 46,
        };
        let time_bytes = timedate.creation_time_bytes().unwrap();
        let date_bytes = timedate.creation_date_bytes().unwrap();
        let mut result = [0u8; 4];
        result[0..2].copy_from_slice(&date_bytes[0..2]);
        result[2..4].copy_from_slice(&time_bytes[0..2]);
        let reconstructed = TimeDate::from_bytes(&result).unwrap();
        assert_eq!(timedate, reconstructed);
    }

    #[test]
    fn test_creation_date_bytes_valid() {
        let date = TimeDate {
            year: 1983,
            month: 04,
            day: 23,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let bytes = date.creation_date_bytes().unwrap();
        assert_eq!(bytes[0], 0x06);
        assert_eq!(bytes[1], 0x97);
    }

    #[test]
    fn test_creation_date_bytes_year_out_of_range() {
        let date = TimeDate {
            year: 1970,
            month: 12,
            day: 31,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let result = date.creation_date_bytes();
        assert!(result.is_err());
        assert_eq!(
            result.unwrap_err().to_string(),
            "Year out of range for MS-DOS."
        );
    }

    #[test]
    fn test_creation_date_bytes_invalid_month() {
        let date = TimeDate {
            year: 1990,
            month: 13,
            day: 31,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let result = date.creation_date_bytes();
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().to_string(), "Invalid month or day.");
    }

    #[test]
    fn test_creation_date_bytes_invalid_day() {
        let date = TimeDate {
            year: 1990,
            month: 12,
            day: 32,
            hour: 0,
            minute: 0,
            second: 0,
        };
        let result = date.creation_date_bytes();
        assert!(result.is_err());
        assert_eq!(result.unwrap_err().to_string(), "Invalid month or day.");
    }
}
