use clap::{Parser, Subcommand};
use disk::Disk;

use crate::{analyzer::Analyzer, builder::DiskBuilder};

pub mod analyzer;
pub mod builder;
pub mod disk;
pub mod downloader;
pub mod fat;
pub mod manifest;
pub mod os;

#[derive(Parser)]
#[command(version="v0.0.3 'Smeagol'", author="Bas v.d. Wiel <bas@area536.com>", about="Virtual DOS hard disk generator.", long_about = None)]
#[command(propagate_version = true)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Build a DOSContainer manifest into a VHD disk image.
    Build {
        /// Filename of the build manifest.
        name: String,
    },
    /// Analyze the contents of a VHD hard drive image.
    Analyze {
        /// Filename of a VHD image file to analyze.
        name: String,
    },
}

fn main() {
    let cli = Cli::parse();
    match &cli.command {
        Commands::Build { name } => {
            // Load the manifest struct from a YAML file
            let manifest = manifest::Manifest::load(name.as_str()).unwrap();
            // Display the application header
            print_header();
            // Display some misc metadata from the manifest we got
            println!("{}", manifest.get_metadata());
            println!("{}", manifest);
            // Construct a DiskBuilder struct from the manifest.
            let mut builder = DiskBuilder::new(manifest).expect("Failed to create a DiskBuilder for this manifest.");
            // If that worked, we're ready to do some actual building.
            builder.build().expect("Failed to build the disk image.");
        }
        Commands::Analyze { name } => {
            print_header();
            let mut analyzer = Analyzer::load(name.as_str()).unwrap();
            analyzer.print_metadata();
        }
    }
}

/// Make the output look like the Bash version.
fn print_header() {
    println!(
        "DOSContainer v0.0.3 disk image toolbox.\n(C)2024 - Bas v.d. Wiel <bas@area536.com>\n"
    );
}
