use std::{fmt, fs::File, io::Read};

use serde::{Deserialize, Serialize};

#[derive(Debug, Deserialize, Serialize)]
pub struct GameMetadata {
    title: String,
    publisher: String,
    year: u32,
    comment: Option<String>,
}

impl fmt::Display for GameMetadata {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Title            : {}\nPublisher        : {}\nYear             : {}\n",
            self.title, self.publisher, self.year
        )
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Application {
    url: String,
    checksum: Option<String>,
    label: Option<String>
}

impl Application {
    pub fn get_url(&self) -> String {
        self.url.to_owned()
    }

    pub fn get_label(&self) -> String {
        if self.label.is_some() {
            self.label.clone().unwrap()
        }
        else {
            self.url.to_owned()
        }
    }
}

#[derive(Debug, Deserialize, Serialize)]
struct GameConfig {
    autoexec: String,
    config: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct OsSettings {
    full_install: Option<bool>,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Manifest {
    pub version: Option<u32>,
    pub metadata: GameMetadata,
    pub diskname: String,
    pub disksize: u32,
    #[serde(default = "default_disktype")]
    pub disktype: String,
    pub application: Vec<Application>,
    pub os: String,
    pub os_settings: Option<OsSettings>,
    #[serde(rename = "AUTOEXEC")]
    pub autoexec: Option<String>,
    #[serde(rename = "CONFIG")]
    pub config: Option<String>,
}

fn default_disktype() -> String {
    "ide".to_string()
}

impl fmt::Display for Manifest {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Diskname         : {}\nDisk size        : {}MB.\nDisk type        : {}\nOperating System : {}\n",
            self.diskname,
            self.disksize,
            Manifest::display_disktype(&self.disktype),
            Manifest::display_os(&self.os)
        )
    }
}

impl Manifest {
    pub fn load(yaml_path: &str) -> Result<Manifest, std::io::Error> {
        let mut file = File::open(yaml_path)?;
        let mut yaml = String::new();
        file.read_to_string(&mut yaml)?;
        let manifest: Manifest =
            serde_yaml::from_str(&yaml).expect("Failed to convert YAML to Manifest.");
        Ok(manifest)
    }

    pub fn get_metadata(&self) -> &GameMetadata {
        &self.metadata
    }

    pub fn default_disktype() -> String {
        "ide".to_string()
    }

    fn display_disktype(disktype: &str) -> String {
        match disktype {
            "pcxt" => "IBM XT".to_string(),
            "pcat" => "IBM AT".to_string(),
            "ide" => "Generic IDE/ATA".to_string(),
            _ => "Unknown disk type".to_string(),
        }
    }

    fn display_os(os: &str) -> String {
        match os {
            "MSDOS200" => "MS-DOS 2.00".to_string(),
            "MSDOS400" => "MS-DOS 4.00".to_string(),
            "MSDOS401" => "MS-DOS 4.01".to_string(),
            "MSDOS500" => "MS-DOS 5.00".to_string(),
            "MSDOS600" => "MS-DOS 6.00".to_string(),
            "MSDOS621" => "MS-DOS 6.21".to_string(),
            "MSDOS622" => "MS-DOS 6.22".to_string(),
            "IBMDOS200" => "IBM PC-DOS 2.00".to_string(),
            _ => "Unknown operating system".to_string(),
        }
    }
}
